/* configutil.c   functions to read configuration files
    Copyright (C) 2000  The Linux Card Emulator Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    http://www.fsf.org/
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#ifndef _WIN32
#include <unistd.h>
#endif
#include <sys/stat.h>
#include "confutil.h"
//#include "ec_keys.h"
#include "debug.h"

char * skip_white_space(char *cp) {
  while ( isspace(*cp) ) cp++;
  return cp;
}

char * skip_separator(char *cp) {
  while ( isspace(*cp) ? 1 : *cp==',' ) cp++;
  return cp;
}

char * skip_chars(char *cp, char *charset) {
  while ( strchr(charset, *cp) ) cp++;
  return cp;
}


/* read config file system:

    ReadConfigFile( char *filename,  ProcSection psect, ProcVariable pvar, 
                    short system);
1)
  define :
    int Section(char *section);
    int VarVal(char *var, char *val, short system);
2)
  while ReadConfigFile() reads the file, 
    Section(section) is called for each section encountered
    VarVal(var,val,system)  is called for each [var] = [val] 
    the equal sign "=" is optional.
    A special command is "#;" that saves the comments for
    each section.
3)
  if no errors are encountered, ReadConfigFile() returns zero.
*/  
  

static void errormsg(int lineno, char *inpline, char *filename,
		     char *fmt, ... ) {
  va_list args;
  msgprint("%3d: %s\n   ** %s: ",lineno,inpline,filename);
  va_start(args,fmt);
  msgprint(fmt,args);
  va_end(args);
}

static int getfilestat(struct stat *pstat, char *fname) {
  int res,err;
  res=stat(fname, pstat);
  err=errno; errno=0;
  if (res) return err;
  else return 0;
}

int file_exist_rd(char *fname) {
  /* check if file exits and has read access */
  struct stat sbuf;
  int errn;
  if ( (errn = getfilestat(&sbuf,fname))==0 ) 
   if (! (S_ISREG(sbuf.st_mode) && (sbuf.st_mode & S_IRUSR))) errn=EACCES;
  return errn ? 0 : 1;
}

int dir_exist(char *dirname, int *err) {
  /* check if dir exits and has apropriate access */
  struct stat sbuf;
  int errn;
  char *dname=dirname;

  if (dname ? strlen(dname)==0 : 1) dname=".";
  if ( (errn = getfilestat(&sbuf,dname))==0 ) {
    if (! S_ISDIR(sbuf.st_mode) ) errn= ENOTDIR;
    else if ( !(sbuf.st_mode & S_IXUSR) ) errn=EACCES;
    else if ( !(sbuf.st_mode & S_IWUSR) ) errn=EACCES;
  }
  /* msgprint("dir_exist(%s)=%d\n",dname,errn); */
  if (err) *err= errn;
  return errn ? 0 : 1;
}

int dir_check(char **dir, char *val) {
  /* return true if "val" is valid and has user access  
     alloc memory to *dir and copy "val" to *dir.
   */
  int ok=0, err;
  char *valstr;

  valstr= SetString_notrsp(val);
  /*if (*dir) free(*dir);*/
  if (strlen(valstr)==0) valstr=".";
  if (dir_exist(valstr,&err)) {
    *dir= calloc(strlen(valstr)+5, sizeof(char));
    if (*dir) { 
      strcpy(*dir, valstr);
      ok=1;
    } else {
      /*msgprint("%3d: %s = %s\n",dconfig_lineno,command,val);*/
      msgprint("   ** internal program error in data_dir=calloc()\n");
    }
  } else {
    /*msgprint("%3d: %s = %s\n",dconfig_lineno,command,val); */
    msgprint("   ** Error, %s : \"%s\"\n",strerror(err),valstr);
  }
  return ok;
}

char *alloc_cat_dir_file(char *dir, char *filename) {
  /* alloc memory for dir & filname, and copy 
     if filename starts with "/", it is an absolute path
  */
  char *fn;
  int dirlen= dir ? strlen(dir) : 0;
  fn=calloc(dirlen+1+strlen(filename)+1,sizeof(char));
  if (fn) {
    if (*filename=='/') strcpy(fn,filename);
    else {
      if (dir) {
	strcpy(fn,dir);
        strcat(fn,"/");
      }
      strcat(fn,filename);
    }
  } else {
    msgprint("   ** internal program error alloc_cat_dir_file(%s,%s);\n"
	     ,dir,filename);
  }
  return fn;
}

int dev_ok(char *fname) {
  struct stat sbuf;
  int sno;
  sno= getfilestat(&sbuf,fname);
  if ( sno==0 ? S_ISCHR(sbuf.st_mode) : 0 ) return 1;
  else return 0;
}

char * find_rc_file(char *rc_name, char *rc_home_name) {
  /* use the first of follwing configuration files:
     1) [rc_name] in current directory
     2) $HOME/.[rc_home_name] 
        if rc_home_name is NULL, use  $HOME/.[rc_name]
  */
  char *home;
  char *rc=NULL;
  /* is the file "lincardemurc"  in current directory  ? */
  if ( file_exist_rd(rc_name) ) return rc_name;
  home = getenv("HOME");
  /* check $HOME/.lincardemurc */
  if (home != NULL) {
    int sz = strlen(home);
    if (rc_home_name) sz += strlen(rc_home_name);
    else sz += strlen(rc_name)+1; /* add "." */
    rc=calloc(sz+2,sizeof(char));
  }
  if (rc) {
    strcpy(rc,home); 
    strcat(rc,"/");
    if (rc_home_name) strcat(rc,rc_home_name);
    else { 
      strcat(rc,".");
      strcat(rc,rc_name);
    }
    if (file_exist_rd(rc)) return rc;
    free(rc);
  } 
  return NULL;
}

/* psect & pvar returns zero if no errors */

int config_lineno=0; /* global , also used in config.c */

int ReadConfigFile(char *fname, ProcSection psect, ProcVariable pvar,
		   short system) {
  /* Read config file, and return number of errors */
  FILE *config;
  int errcnt=0;
  int preverr=0;
  char section[COMMANDSZ];
  char inpline[LINESIZE];
  char command[COMMANDSZ];

  if((config = fopen(fname, "r")) == NULL) {
    errprint(errno,"Couldn't open config file: %s\n",fname);
    return -1;
  }
  /* msgprint("using configuration file %s\n",fname); */
  config_lineno=0;
  errno=0;
  while (fgets(inpline,LINESIZE,config)) {
    char *lnp;

    /* remove EOLN : */
    if ((lnp=strchr(inpline,'\n'))) *lnp='\0';
    if ((lnp=strchr(inpline,'\r'))) *lnp='\0';
    
    config_lineno++;
    lnp= skip_white_space(inpline);
    /* skip comments : */
    if (*lnp==';'|| *lnp=='\r'|| *lnp=='\n'|| *lnp=='\0') continue;
    if (*lnp=='#') if ( *(lnp+1) != ';' ) continue;

    /* DEBUGPR(0xff,("%d: %s",config_lineno,lnp)); */
    *section='\0';
    if ( !strncmp(lnp,"#;",2) ) {
      /* save / process  comments : */
      if (pvar("#;",lnp+2,system)) {
	errcnt++;
	errormsg(config_lineno,inpline,fname,
		 "error in comment, or comment not allowed here\n");
	preverr=errcnt;
      }
    } else if (strcmp(lnp,"[]")? (sscanf(lnp,"[%31[^]]]", section)==1) : 1) {
      if ( psect(section,system)) {
	errcnt++; 
	errormsg(config_lineno,inpline,fname,
		 "unknown section: [%s]\n", section);
	preverr=errcnt;
	}
    } else if ( sscanf(lnp, "%31[^ \t=]", command)!=1   ) {
      errcnt++; /* syntax error */
      errormsg(config_lineno,inpline,fname,
	       "syntax error, expected:  command [[=] value]\n");
      preverr=errcnt;
    } else {
      int l;
      l= strcspn(lnp,"= \t");
      lnp= &lnp[l];
      l= strspn(lnp,"= \t");
      lnp= &lnp[l];
      /* printf("...%s\n",lnp); */
      l=strcspn(lnp,"#;\r\n");
      lnp[l]='\0';
      if (pvar(command,lnp,system)) {
	errcnt++;
	errormsg(config_lineno,inpline,fname,
		 "syntax error in assignment value\n");
	preverr=errcnt;
      }
    }
    if (preverr<errcnt ) {
      if (errno) 
	errprint(errno,"Error reading config file\n","ab") ;
      errormsg(config_lineno,inpline,fname,"syntax error\n");
      preverr=errcnt;
      errno=0;
    }
  }
  if (errno) {
    errprint(errno,"Errors reading config file\n","ab") ;
    errcnt++;
  }
  if (errcnt)
    msgprint("There were %d errors reading %s\n",errcnt,fname) ;
  /* DEBUGPR(0xf,("%s: %d lines\n",file,config_lineno));*/
  return   errcnt;
}
/*                                   // vyhadzuje warning
void add_comment(comment_rec **first, char *val) {
  comment_rec *newptr, *ptr;
  newptr= (comment_rec *) malloc( sizeof(comment_rec) );
  newptr->next=NULL;
  if (val) {
    newptr->text = (char *) calloc(strlen(val)+1, sizeof(char));
    strncpy(newptr->text, val, strlen(val)+1);
  } else
    newptr->text=NULL;
  if (! *first) *first= newptr;
  else {
    for (ptr=*first; ptr->next; ptr= ptr->next) ;
    ptr->next= newptr;
  }
}
*/
void print_comments(FILE *keyfile, comment_rec *comment_ptr) {
  if (comment_ptr) {
    comment_rec *ptr;
    for (ptr=comment_ptr; ptr; ptr=ptr->next)
      if (ptr->text) fprintf(keyfile,"#;%s\n",ptr->text);
  }
}

void str_eatspaces(char *p) {
  /* remove  spaces and other garbage from string 
     "  abc def  "  ->  "abc def"
  */
 unsigned char *d,*s;
 if (p==NULL || *p == '\0' ) return;
 d=s= (unsigned char *) p; 
 while (*s != '\0') 
        if (!isspace(*s)) *d++=*s++;
        else s++;
 *d='\0'; 
}


void str_notrspaces(char *dst) {
  /* remove leading and trailing spaces in string  */
  unsigned char *d;
  unsigned char *s1,*s2;

  if (dst==NULL || *dst == '\0' ) return;
  d=s1= (unsigned char *) dst; 
  while (*s1 != '\0' && isspace(*s1) )  s1++;
  s2=strlen(s1)+s1;
  if (s2>s1) {
    --s2;
    while (*s2 != '\0' && isspace(*s2) )  *(s2--)='\0';
  }
  if (d<s1) {
    while (*s1 != '\0') { *d++ = *s1++; }
    *d='\0';
  }
}


int StrNoCaseCmp(const char *s1, const char *s2)
{ /* as strcmp() except case insensitive */
   int rs;
   if (s1==NULL && s2==NULL) return 0;
   else if (s1==NULL) return -1;
   else if (s2==NULL) return 1;
   while ( 1 ) {
     if ( *s1=='\0' && *s2=='\0' ) return 0;
     else if (*s1=='\0') return -1;
     else if (*s2=='\0') return 1;
     rs=tolower(*s2++)-tolower(*s1++);
     if (rs != 0) return rs;
   }
   return 0;
}

int StrNoCaseCmpN(const char *s1, const char *s2, unsigned int n)
{ /* as strcmpn() except case insensitive */
   int rs;
   if (s1==NULL && s2==NULL) return 0;
   else if (s1==NULL) return -1;
   else if (s2==NULL) return 1;
   while ( n>0 ) {
     if ( *s1=='\0' && *s2=='\0' ) return 0;
     else if (*s1=='\0') return -1;
     else if (*s2=='\0') return 1;
     rs=tolower(*s2++)-tolower(*s1++);
     if (rs != 0) return rs;
     n--;
   }
   return 0;
}


char *SetString(char *val) {
  /* modify "val" to be a correct string , string must be R/W  */
  int n;
  char *str=val;
  if (val==NULL) return "";
  if (strlen(val)==0) return val;
  str= skip_white_space(str);
  if (*str=='"') n= strcspn(++str, "\"\n\r");
  else n= strcspn(str, "\n\r");
  if (str[n]!='\0') str[n]='\0';
  return str;
}

char *SetString_notrsp(char *val) {
  /* modify "val" to be a correct string with no leading/trailing spaces etc.
     " abc.txt "      => abc.txt
        my.txt   \n  =>  my.txt
  */
  val=SetString(val);
  str_notrspaces(val);
  return val;
}

u8 MM2toHex(unsigned mm2, unsigned *pos, short key_length) {
/*  decimal multimackode: ABCD 
    D=position saved in *pos ,  position 0..(key_length-1)
    is C>3  then *pos = D+KEY_LENGTH  : usually 8..15  (ECS3)
    if error : pos:= 0xffnn
*/
  u16 h,ABC,AB,C,A,B;
  h= mm2%10;
  ABC = mm2/10;
  C= ABC%10;
  AB = ABC/10;
  B=AB%10;
  A=AB/10;
  /* catch some non-valid MM2 codes */
  if (h>=key_length) *pos=0xff1f;
  else if (C>7) *pos=0xff2f;
  else if (B<1||B>8) *pos=0xff4f;
  else if (A<1||A>8) *pos=0xff8f;
  else *pos=h;
  h= (((A-1)<<4)&0xf0) + B-1;
  /*printf("mm2: %u b: %d-> hex: %02x\n",ABCD,b,h);*/
  h += C & 1 ? 0x8 : 0;
  h += C & 2 ? 0x80 : 0;
  /* printf("mm2: %u -> hex: %02x ",mm2,h);*/
//  if ( C > 3) *pos += KEY_LENGTH;
  return h;
}

unsigned HextoMM2(u8 H, unsigned pos) {
  /* pos 0..7 or 8..15 */
  unsigned mm2;
  u8 b,h;
  H &= 0xff;
  b = (H&0xf) >= 0x8 ? 1 : 0;
  b += H>=0x80 ? 2 :0;
  h = (H&0x77) + 0x11;
  mm2 = (((h>>4)*10 + (h&0xf))*10 + b)*10;
 // mm2 += (pos >=MM2KEY_LENGTH) ? (pos-MM2KEY_LENGTH)+40 : pos;
  /* printf("hex: %02x -> mm2: %04u ",H,ABCD);*/
  return mm2;
}

static char *SetU_valp=NULL;

int SetUL(char *val,unsigned long *var, unsigned sz, int base) {
  /* return 1 if ok, 0 if error */
  unsigned long v;
  char *cp;
  v=strtoul(val, &cp, base);
  if (sz<4) v &= (~0UL)>>(32-sz*8);
  if (cp<=val) return 0;
  SetU_valp=cp;
  *var=v;
  return 1;
}

int SetU(char *val,unsigned *var, unsigned sz, int base, 
	 unsigned min, unsigned max) {
  /* return 1 if ok, 0 if error */
  unsigned long v;
  int r;
  v = *var;
  r=SetUL(val, &v, sz, base);
  if (r)  *var = (unsigned) v;
  if (*var<min || *var>max) r=0;
  return r;
}

char *HextoChar="0123456789abcdef";

/*
u8 ChartoHex(char c) {
  char *p;
  p=strchr(HextoChar, tolower(c));
  if (!p) return 0xff;
  return p-HextoChar;
}
*/

int SetCHMASK8(char **val, u8 *var,u8 *mask,unsigned sz) {
  char *cp,*hexp;
  u8   *v,*m;
  int cnt;
  *val= skip_white_space(*val);  
  v=var; m=mask;
  for (cnt=0; cnt<sz/2; cnt++) {
    *(v++)= 0x00;
    *(m++)= 0xff;
  }
  v=var; m=mask;
  for (cp=*val; (cp - *val)<sz; cp++) {
    u8 mv;
    if ((cp - *val)&1) mv=0x0f;
    else mv=0xf0;
    if (*cp=='*') {
      *m &= ~mv;
    } else if ((hexp=strchr(HextoChar, tolower(*cp)))) {
      u8 h;
      h= mv&0xf0 ? (hexp-HextoChar)<<4 : hexp-HextoChar;
      *v |= mv & h;
    } else break;
    if (mv&0x0f) { v++; m++; }
  } 
  cp= skip_white_space(cp);
  if (cp<= *val+1) return 0;
  *val=cp;
  return 1;
}

int SetCHMASK(char **val, u32 *var,u32 *mask,unsigned sz) {
  char *cp,*hexp;
  *var=0; *mask=0xffffffffUL;
  *val= skip_white_space(*val);  
  for (cp=*val; (cp - *val)<sz; cp++) {
    if (*cp=='*') {
      *var  <<= 4;
      *mask <<= 4;
      *mask &= 0xfffffff0UL;
    } else if ((hexp=strchr(HextoChar, tolower(*cp)))) {
      *var  = ((*var  <<4) & 0xfffffff0UL) | (hexp-HextoChar) ;
      *mask = ((*mask <<4) & 0xffffffffUL) | 0xfUL;
    } else break;
  } 
  cp= skip_white_space(cp);  
  if (cp<= *val+1) return 0;
  *val=cp;
  return 1;
}

int SetU32(char *val, u32 *var, u32 *pskip,unsigned sz, int base, u32 min, u32 max) {
  /* sz= size of val:  val[0..sz-1]
     return the number of values read from string 
     if hex_value == '%' is undefined/skip, but accepted as ok
  */
  char *cp;
  char lval[2];
  u32 skip=0,bitset=1;
  int  vi, res=0;
  unsigned long hx=0;
  lval[1]='\0';
  cp=val;

  for (vi=0; vi<sz; ++vi) var[vi]=0;
  vi=0;
  if (base==10||base==0||base==8) {
    for (; vi<sz; ++vi) {
      cp= skip_separator(cp);
      if (isdigit(*cp)) {
	hx=strtoul(cp,&cp,base);
	++res;
      } else break;
      if (vi<sz) var[vi]=  hx;
      /* else do not store value with error */
      if (hx>max || hx<min) break;
    }
  } else if (base==16) {
    for (; vi<sz; ++vi) {
      cp= skip_separator(cp);
      if (*cp=='%' && pskip) {
	cp++;
	hx=min;
	if (vi<sz) {
	  skip |= bitset;
	}
	++res;
      } else if (isxdigit(*cp)) {
	hx=strtoul(cp,&cp,16);
	++res;
      } else break;
      if (vi<sz) var[vi]= hx;
      /* else do not store value with error */
      if (hx>max || hx<min) break;
      bitset += bitset; /* shift left one */
    }
  }
  res=  vi;
  SetU_valp=cp;
  if (pskip) *pskip= skip;
  return res;
}

u8 nord4_ID[3], nord4_key;
char *SetU8_valp=NULL;

int SetU8(char *val, u8 *var, unsigned sz, int base, int min, int max) {
  /* sz= size of val:  val[0..sz-1]
     return the number of values read from string */
  char *cp;
  char lval[2];
  int  vi, res=0;
  unsigned long ul;
  unsigned hx=0;
 // u16  nord4[NORDIC4_LENGTH];
  lval[1]='\0';
  cp=val;
  for (vi=0; vi<sz; ++vi) var[vi]=0;
  vi=0;
  if (base==16 ) {
    for (; vi<sz; ++vi) {
      cp= skip_separator(cp);
      if (isxdigit(*cp)) {
	*lval = *cp;
	ul=strtoul(lval,NULL,16);
	hx = ul  ;
	++cp;
	++res;
      } else break;
      if (isxdigit(*cp)) {
	*lval = *cp;
	ul=strtoul(lval,NULL,16);
	hx = (hx << 4);
	hx |= ul & 0x0f  ;
	++cp;
	++res;
      } ;
      var[vi]= (u8) hx; 
      if (hx>max || hx<min) break;
    }
  } else  if (base==10 || base==MM2base || base==Nordic4base) {
    for (; vi<sz; ++vi) {
      unsigned pos=vi;
      cp= skip_separator(cp);
      if (isdigit(*cp)) {
	ul=strtoul(cp,&cp,10);
	hx = ul  ;
	++res;
      } else break;
      if (base==MM2base) {
	hx= MM2toHex(hx, &pos, sz);
	/* if (pos>=sz) pos=vi; if error skip... */
	if (pos<sz) { 
	  var[pos]= (u8) hx;
	} /* else do not store value with error */
      } else if (base==Nordic4base ) {
//	if (vi<NORDIC4_LENGTH) nord4[vi]= hx;
	/* else do not store value with error */
      } else /* std. dec */ {
	if (pos<sz) var[pos]= (u8) hx;
	/* else do not store value with error */
      }
      if (hx>max || hx<min) break;
    }
    if (base==Nordic4base) {
   //   Nordic4toHex(var,nord4_ID,&nord4_key,nord4);
    }
  }
  res=  vi;
  SetU8_valp=cp;
  return res;
}

void keys_to_string(char *s, byte n, byte keys[]) {
  int ik;
  if (n>0) sprintf(s,"%02x", keys[0]);
  s +=2;
  for (ik=1; ik<n; ik++,s +=3)
    sprintf(s," %02x", keys[ik]);
}



int SetDateTime(char *val, datetime *dt) {
  /* return 1 if ok date, 3 if ok date+time, 0 if error */
  char skip[22];
  int year,mon,day,hour,min,sec;
  int n,rs=0;
  /* printf("[%s] ",val); */
  n=sscanf(val,"%4d%1[-/. ]%2d%1[-/. ]%2d%[, ]%2d%1[:. ]%2d%1[:. ]%2d",
	      &year,skip,&mon,skip,&day,skip,
	      &hour,skip,&min,skip,&sec);
  if (n>=5) {
    dt->year=year; dt->month=mon; dt->mday= day;
    rs=1;
  } 
  if (n>=11) {
    dt->hour= hour; dt->min= min; dt->sec= sec;
    rs=3;
  }
  return rs;
}


