/* debug.c - functions for debug output and loging to files Copyright
    (C) 2000-2002 The Linux Card Emulator Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    http://www.fsf.org/ */

#include "common.h"
#ifdef USE_WIN32
#include <windows.h>
#else
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#ifdef USE_WIN32
#include <conio.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "debug.h"

/* from confutil.c : */
extern char *alloc_cat_dir_file(char *dir, char *filename) ;

#ifdef USE_NCURSES
WINDOW *logwin, *commwin,*statwin;
#else
int  *logwin=NULL, *commwin=NULL, *statwin=NULL;
#endif

/*
#define SER_DEBUG 0x100f
#define SER_DEBUG 0x0000
*/


int serdebugval[4] = { 0x0, 0x3, 0x7, 0x100f};
int stddebugval[4] = { 0x1, 0x3, 0x7, 0xf};

int serdebug = 0x1 ;

loginfo *mainlog;
loginfo *updlog;


#ifndef USE_WIN32
static struct termios saved_termios_data;
#endif

void tty_inputmode(int on) {
  /* 0: no line buffering of input => no ICANON !! and no echo  */
  /* 1: restore from saved_termios_data */
#ifndef USE_WIN32
  int fd=1;
  if (on) {
    tcsetattr ( fd, TCSAFLUSH, &saved_termios_data );  
    /* alternatives: TCSADRAIN and TCSANOW */
  } else /* off */ {
    static struct termios termios_data;
    tcgetattr ( fd , &saved_termios_data );
    termios_data= saved_termios_data;
    termios_data.c_cc[VMIN]  = 1;
    termios_data.c_cc[VTIME] = 0;
    termios_data.c_lflag &= ~(ECHO | ICANON);
    tcsetattr ( fd, TCSAFLUSH, &termios_data );  
  }
#endif
}


/* get character from keyboard/stdin */
int scrgetch() {
  int ch;
#ifdef USE_WIN32
  /* skip extended keycodes (arrows, etc...) */
  do {
    ch= getch();
    if (ch==0||ch==0xe0) {    ch=getch(); ch= -2;   }
  } while (ch==-2);
#else
#ifdef USE_NCURSES
  if (commwin) {
    wprintw(commwin,"lincardemu>");
    ch= wgetch(commwin);
    wprintw(commwin,"\r           \r",ch);
  } else 
#endif
    ch= getchar();
#endif
  return ch;
}

int curr_serdebug() { return serdebug; };

void debug_ser_set(int v) {
  serdebug |= v;
}
void debug_ser_clear(int v) {
  serdebug &= ~v;
}

int debug_ser_toggle() {
  static int togval=0;
  togval = (togval+1)%4;
  serdebug= serdebugval[togval];
  return togval;
}
int debug_toggle() {
  static int togval=0;
  togval = (togval+1)%4;
  mainlog->debugval= stddebugval[togval];
  updlog->debugval= stddebugval[togval];
  return togval;
}

void set_new_loginfo(loginfo *mlog, loginfo *ulog) {
  mainlog=mlog;
  updlog=ulog;
  if ( commwin ) {
    /* assume curses on */
    mlog->out = To_win;
    ulog->out = To_win;
  }
}

#if 0
void debug_set_stderr(int on) {
  /* set debug to stderr on/off */
  mainlog->out &=  ~ To_stderr;
  if (on) mainlog->out |=  To_stderr;
}
void debug_set_stdout(int on) {
  /* set debug to stdout on/off */
  mainlog->out &=  ~ To_stdout;
  if (on) mainlog->out |=  To_stdout;
}
#endif

void set_logfilename(loginfo *linf, char *flname) {
  /* if flname is non empty the log file *flname is opened in append mode
     if flname is the empty string, an open log file is closed */
  /* close open log file */
  if (linf->log_file_name) free(linf->log_file_name);
  linf->log_file_name=NULL;
  linf->log_file_name=alloc_cat_dir_file(log_dir, flname);
  /* linf->log_file_name=calloc(strlen(flname)+1,sizeof(char));
     strcpy(linf->log_file_name,flname); */
}

int have_logfilename(loginfo *l) {
  return l ? (l->log_file_name ? *(l->log_file_name) : 0) : 0;
}


char *current_logfilename(loginfo *linf) {
  if (!linf) return "";
  if (have_logfilename(linf) && linf->logfile) return linf->log_file_name;
  else return NULL;
}


static char *log_destinations[]=
{ "none", "stdout","window","stdout & window",
  "log file","stdout & log file", 
  "window & log file", "stdout & window &log file"
};

char * get_curr_loginfo(loginfo *linf){
  int l;
  l= linf? (linf->logfile ? linf->out : linf->out&~To_logfile) : 0;
  return log_destinations[l & 0x7];
}

void scrprint_loginfo(loginfo *linf) {
  char *fname="";
  char *txt="unknown file";
  if (linf==mainlog) txt= "main data";
  else if (linf==updlog) txt="updates";
  if (linf->logfile) fname=linf->log_file_name;
  scrprint("Log %s to : %s (%s%s)\n", txt,
	   get_curr_loginfo(linf),  (*fname? "file=" : ""),fname);
}

void  toggle_log(loginfo *linf) {
  int skipmask=To_stderr;
  if (logwin) skipmask |= To_stdout;
  else skipmask |= To_win;
  /* toggle different log values, skip values with stderr */
  do {
    linf->out = (linf->out+1) & (To_stdout|To_win|To_logfile);
  } while (linf->out&skipmask) ;
}

void close_log(loginfo *linf) {
  if (linf->logfile)  fclose(linf->logfile);
  linf->logfile=NULL;
}

void flush_win() {
#ifdef USE_NCURSES
  if (logwin) wrefresh(logwin);
  if (commwin) wrefresh(commwin);
#else
  fflush(stdout);
#endif
}

void flush_logs() {
  if (mainlog->logfile)  fflush(mainlog->logfile);
  if (updlog->logfile)  fflush(updlog->logfile);
#ifdef USE_NCURSES
  flush_win();
#else 
  fflush(stdout);
#endif
}

void open_log(loginfo *linf) {
  /* open logfile in append mode */
  close_log(linf);
  if ( have_logfilename(linf) ) {
    /* open file for append */    
    if (! (linf->logfile=fopen(linf->log_file_name,"a"))) 
      errprint(errno," error with file %s\n",linf->log_file_name);
  }
}

int window_width() { 
#ifdef USE_NCURSES
  return COLS; 
#else
  return 255;
#endif
}

/* use for status line with ncurses */
void statprint(char *fmt, ...) {
#ifdef USE_NCURSES
  va_list args;

  va_start(args,fmt);
  if (statwin) {
    vwprintw(statwin,fmt,args);
    wrefresh(statwin);  
  }
  va_end(args);
#endif
}


#ifdef USE_NCURSES
/* use main log output/stdout settings , unconditional print */
void scrprint(char *fmt, ...)
{
  va_list args;

  va_start(args,fmt);
  if (commwin) vwprintw(commwin,fmt,args);
  else 
    vfprintf(stdout,fmt,args);
  va_end(args);
}

#endif

/* print format+data to output/logfile according to loginfo data (linf) */
void vdbprint(loginfo *linf, loginfo *altl, char *fmt, va_list args)  {
  if (linf? linf->out & To_stderr: 1) vfprintf(stderr,fmt,args);
  if (linf->out & To_stdout) vfprintf(stdout,fmt,args);
#ifdef USE_NCURSES
  if ((linf->out & To_win) && logwin) vwprintw(logwin,fmt,args);
  if ((linf->out & To_commwin) && commwin) vwprintw(commwin,fmt,args);
#endif
  if ((linf->out&(Skip_logfile|To_logfile))==To_logfile) {
    if (linf->logfile ) {
      /* only if logfile is open */
      vfprintf(linf->logfile,fmt,args);
      
    } 
    if (altl ? altl->logfile /*&& !have_logfilename(linf)*/  : 0 ) {
      /* also want output to secondary logfile, so : */
      vfprintf(altl->logfile,fmt,args);
    }
  }
}

/* use main log output/stdout settings , conditional
   only if (lval & mainlog->debugval) : */
void mlogprint(int debval, char *fmt, ...) {
  loginfo *log1, *log2;
  if (debval&DEB_UPDLOG ? updlog->logfile : 0) {
    /* only if upd-logfile is open ! */
    log1=updlog; log2=mainlog;
  } else {
    log1=mainlog; log2=NULL;
  }
  if (debval & log1->debugval ) {
    va_list args;
    unsigned char out;

    va_start(args,fmt);
    out= log1->out;
    if (debval&DEB_NOLOG) log1->out |= Skip_logfile;
    if (debval&DEB_SCR) log1->out |= commwin ? To_commwin : To_stdout;
    if (debval&DEB_ONLYLOG) log1->out &= ~To_console;

    vdbprint(log1,log2,fmt,args);
    log1->out = out;
    va_end(args);
  }
}

void mlogprint_buffer(int dbg, unsigned char *buf, int first, int sz,
			     char *msg) {
  int cnt,l=0;
  mlogprint(dbg,"%s",msg);
  for (cnt=first; cnt<sz; cnt++,l++)  {
    if (l>0 && l%20==0) mlogprint(dbg,"\n\t;");
    mlogprint(dbg," %02x",buf[cnt]);
  }
}

#if 0
/* use update log output/stdout settings , conditional
   only if (lval & updlog->debugval) 
   if no update log file, use main log file      */
void ulogprint(int debval, char *fmt, ...) {
  if (debval & updlog->debugval ) {
    va_list args;
    unsigned char out;

    va_start(args,fmt);
    out= updlog->out;
    if (debval&DEB_NOUPDLOG) updlog->out &= ~To_logfile;
    /*fprintf(stderr,"dbg=%0x out=%x -> %x \n",debval,out, updlog->out);*/
    vdbprint(updlog,mainlog,fmt,args);
    updlog->out = out;
    va_end(args);
  }
}
#endif

/* use main log output/stdout settings , unconditional print */
void dbprint(char *fmt, ...)
{
  va_list args;

  va_start(args,fmt);
  vdbprint(mainlog,NULL,fmt,args);
  va_end(args);
}

/* print to logfile + minimum stderr if stdout is not used ,
   can be used for other (error) messages with no system errno */
void msgprint(char *fmt, ... )
{
  va_list args;
  unsigned char out=0;

  va_start(args,fmt);
  if (mainlog) {
    out=mainlog->out;
    if ( ! (out & To_console)) mainlog->out |= To_stderr;
  }
  vdbprint(mainlog,NULL,fmt,args);
  if (mainlog) mainlog->out = out;
  va_end(args);
}

/* print errormessage and errorstring from "err=errno" with strerror(),
   output to current "mainlog/", and minimum stderr if stdout is not used
   Should be used for error messages : */
void errprint(int err, char *fmt, ...) {
  va_list args;
  unsigned char out=0;
  char *  strerr;

  if (err==104) strerr="Connection reset by peer";
  else if (err>0) {
    /* "109" this is due to error in strerror() on cygnus gcc */
    strerr=strerror(109); /* provoke "unsupported" message */
    if ((strerr=strerror(err))==NULL)
    strerr="No error msg. available";
  } else {
    strerr="";
    err = -err;
  }
  if (mainlog) {
    out=mainlog->out;
    if (!(out & To_console)) mainlog->out |= To_stderr;
  }
  if (err) dbprint("** Error No.= 0x%x (%s)\n\t\t",err,strerr);
  else dbprint("** Error : ");
  va_start(args,fmt);
  vdbprint(mainlog,NULL,fmt,args);
  va_end(args);
  if (mainlog) mainlog->out = out;
}

char time_stamp[32]; /* todays date / timestamp in ascii */
struct tm today_now; /* todays date */

void mlogpstr(int debval,char *str) {
  loginfo *log1, *log2;
  if (debval&DEB_UPDLOG ? updlog->logfile : 0) {
    /* only if upd-logfile is open ! */
    log1=updlog; log2=mainlog;
  } else {
    log1=mainlog; log2=NULL;
  }

  if (debval & log1->debugval) {
    unsigned char out;
    out= log1? log1->out : To_stdout;
    if (debval&DEB_NOLOG) out |= Skip_logfile;
    if (debval&DEB_SCR) log1->out |= commwin ? To_commwin : To_stdout;
    if (debval&DEB_ONLYLOG) out &= ~To_console;
    
    if (out & To_stderr) fputs(str,stderr);
    if (out & To_stdout) fputs(str,stdout);
#ifdef USE_NCURSES
    if ((out & To_win) && logwin) waddstr(logwin,str);
    if ((out & To_commwin) && commwin) waddstr(commwin,str);
#endif
    if (!log1) return ;
    if (log1->logfile && (out & (Skip_logfile|To_logfile))==To_logfile) 
      fputs(str,log1->logfile);
    if (!log2) return;
    if (log2->logfile)  /* try alternate logfile  */
	fputs(str,log2->logfile);
  }
}

#if 0
void ulogpstr(int debval,char *str) {
  if (debval & updlog->debugval) {
    unsigned char out;
    out= updlog? updlog->out : To_stdout;
    if (debval&DEB_NOUPDLOG) out &= ~To_logfile;
    if (debval&DEB_NOLOG) out |= Skip_logfile;
    if (debval&DEB_SCR) mainlog->out |= commwin ? To_commwin : To_stdout;
    if (debval&DEB_ONLYLOG) out &= ~To_console;
    
    if (out & To_stderr) fputs(str,stderr);
    if (out & To_stdout) fputs(str,stdout);
#ifdef USE_NCURSES
    if ((out & To_win) && logwin) waddstr(logwin,str);
    if ((out & To_commwin) && commwin) waddstr(commwin,str);
#endif
    if (!updlog) return ;
    if ((out&(Skip_logfile|To_logfile))==To_logfile) {
      if (updlog->logfile )   /* only if logfile is open */ 
	fputs(str,updlog->logfile);
      else if (mainlog ? mainlog->logfile : 0 )  /* try alternate logfile  */
	fputs(str,mainlog->logfile);
    }
  }
}

#endif

void set_time_stamp() {
  time_t t;
  if (time(&t) == ((time_t)-1)  ) {
    /*error;*/
    strcpy(time_stamp,"----.--.-- --:--:-- ");
  } else {
    today_now = *localtime(&t);
    strftime(time_stamp,31,"%Y.%m.%d %H:%M:%S ",&today_now);
  }
}



#ifdef DONT_HAVE_GETOPT
/* A very simple getopt() */
/* for cygwin32 with mingw32 (-mno-cygwin) or other Windows C compilers */
static char **argvl=NULL;
static char *currarg=NULL;
static int maxargc=-1;
int optind = -1;
char *optarg = NULL;
int  optopt=0,opterr=0;

int getopt(int argc, char *argv[], char *optstring) {
  char *cp;

  if ( optind == -1) {
    maxargc=argc; 
    argvl=argv; 
    optind=0;
  }
  if (currarg==NULL || *currarg=='\0') {
    if (++optind>=maxargc) return -1;
    currarg=argvl[optind];
    if (*currarg != '-') return -1;
    currarg++;
  }
  optopt=*currarg;
  if (*currarg=='\0') return '?';
  cp=strchr(optstring,*(currarg++));
  if (cp ==NULL ) return '?';
  if (*(cp+1)==':') {
    if (*currarg!='\0') { optarg=currarg; currarg=NULL; }
    else {
      if (++optind>=maxargc) return '?';
      optarg=argvl[optind];
      currarg=NULL;
    }
  } else {
    optarg=NULL;
  }
  return *cp;
}

#endif

