#ifndef _DEBUGH
/* debug.h - functions for debug output and loging to files
    Copyright (C) 2000  The Linux Card Emulator Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    http://www.fsf.org/
*/
#define _DEBUGH
#ifdef _WIN32
#undef USE_NCURSES
#endif

#include <stdarg.h>
/* <errno.h> exists in unix and Borland C++ */
#include <errno.h>
#include <time.h>
#include <stdio.h>

#ifdef USE_NCURSES
#include <ncurses.h>
#endif

#undef DONT_HAVE_GETOPT
#ifdef _WIN32
#ifdef __CYGWIN__
#include <getopt.h>
#else
#define DONT_HAVE_GETOPT
#endif
#endif

#ifdef  DONT_HAVE_GETOPT
extern int optind ;
extern char *optarg;
extern int optopt,opterr;
int getopt(int argc, char *argv[], char *optstring) ;
#endif

/*
    When writing to log files and/or stdout/stderr use:

      mlogprint(dbval,..), ulogprint(dbval,..),
      mlogpstr(dbval,..), ulogpstr(dbval,..),
      mlogtimestamp(dbval), ulogtimestamp(dbval)

   There are 2 possible log files, "mainlog" and "updatelog".
   If "updatelog" is not defined in config.dat, mainlog is used
   Updatelog is used if the dbval is OR-ed with DEB_UPDLOG :
        mlogprint(dbval|DEB_UPDLOG,....)

   Different filnames may be used for each system (ec, seca,....),
   the file name is defined in config.dat under each section.

   For writing messages that at least are sent to stderr or stdout :
   : system error messages using "errno" use  errprint(errno,....);
   : other debug/error messages  use  msgprint(...);

   For debug messages to current selected medium use  dbprint(...);
   Instead of dbprint() the macro DEBUGPR(...) may be used, since
   if compiled with NODEBUG all DEBUGPR(...) statements will not generate code

  for writing to "screen" only, use scrprint() == macro for printf()
  of just use stderr/stdout.....
*/

/* debug values:
  least significant bit is for messages that always should be ouput
  that is debug values in order of significance (important first) should be:
     0x1, 0x2, 0x4, 0x8, ....
     "debugval" will ususally have the values 0x1, 0x3, 0x7, 0xf
  DEB_NOLOG:    force NO output to logfile.
  DEB_ONLYLOG:  force not writing to stdout/stderr.
  DEB_UPDLOG:   write also to update log file.
  DEB_SCR :     force writing to either stdout or ncurses window.
ex.
  DEB_NOLOG|0x1 will only send message to stdout/stderr if these are enabled
*/

enum { DEB_ALL=0x07ff, DEB_NOLOG=0x8000, DEB_ONLYLOG=0x4000,
       DEB_UPDLOG=0x2000, DEB_SCR=0x1000  } ;

enum { To_stdout=1, To_win=2, To_logfile=4, To_commwin=8,
       To_stderr=0x10, Skip_logfile=0x20 ,
       To_console= To_stdout|To_win|To_commwin|To_stderr  };


/* use commwin to test if curses command window is active */
#ifdef USE_NCURSES
extern WINDOW *logwin, *commwin, *statwin;
#else
#define A_NORMAL 0
extern int *logwin, *commwin,*statwin;
#endif

typedef struct  loginfo {
  int   debugval;
  FILE  *logfile;
  char  *log_file_name;
  unsigned char  out;
} loginfo ;


extern loginfo *mainlog, *updlog;

int  debug_toggle() ;            /* toggle debug, 4 different... */
int  debug_ser_toggle() ;        /* toggle serial debug , ....   */
#if 0
/* not used , obsolete */
void debug_set_stderr(int on) ;  /* set debug to stderr on/off */
void debug_set_stdout(int on) ;  /* set debug to stdout on/off */
void debug_ser_set(int ) ;       /* set/clear bits, overriding current */
void debug_ser_clear(int ) ;     /* serial debug values */
#endif

/* initialize loginfo, usually when config.dat is read or new "c_system" */
void set_new_loginfo(loginfo *mlog, loginfo *ulog) ;
/* toggle log : none, stdout, logfile, stdout+logfile */
void toggle_log(loginfo *linf) ;
char * get_curr_loginfo(loginfo *linf);

/* if flname is non empty the log file "*flname" is opened in append mode
   if flname is the empty string, an open log file is closed */
void set_logfilename(loginfo *linf, char *flname) ;
void open_log(loginfo *linf);
void close_log(loginfo *linf);
void flush_logs();
void flush_win();
char *current_logfilename(loginfo *linf) ;
int have_logfilename(loginfo *l) ;
void scrprint_loginfo(loginfo *linf) ;

/* scrgetch() returns character (no echo) */
int scrgetch();

/* get whole input line, with some editing facilities,
   if compiled with readline this will be used in non-ncurses-mode */
char * console_getline(char *prompt) ;

/* return current serial debug value */
int curr_serdebug() ;
extern int serdebug;

/* only used when reading config.dat in config.c */
extern int serdebugval[4];
extern int stddebugval[4];

/*
   if compiled with NODEBUG all DEBUGPR(...) will not generate code

   DEBUGPR:
   The "mask"-value is anded with debugval, and if the result is not 0
   a debug message  will be sent.

   SERDEBUG: is similar to DEBUGPR (uses serdebug), and is only used
   for testing serial functions
 examples :

   DEBUGPR( mask, " This is a debug message no. %d, %s\n", no,messg) );
*/


#ifdef NODEBUG
/* no debug code, just dummies.... */
#  define DEBUGPR(db,debugval)
#  define SERDEBUGPR(db,serdebug)
#else
/* debug code is include */
#  define DEBUGPR(db,arglist)  if (mainlog->debugval & (db)) dbprint arglist
#  define SERDEBUGPR(db,arglist)  if (serdebug & (db)) dbprint arglist
#endif

/* used for status line with ncurses , when not using ncurses do nothing... */
void statprint(char *ftm, ...);
/* return max line length that fits on ncurses windows */
int window_width() ;

#ifdef USE_NCURSES
/* scrpint(): use main log output/stdout settings , unconditional print */
void scrprint(char *fmt, ...);
#else
#define scrprint printf
#endif

/* explicit functions (not macros) : */

/* print to main logfile + minimum stderr if stdout is not used ,
   can be used for other (error) messages with no system errno */
void msgprint(char *fmt, ... ) ;

/* Print errormessage and errorstring from "err=errno" with strerror(),
   output to current main logfile + minimum stderr if stdout is not used.
   Should be used for system error messages , example:
     errprint(errno, "Additional error %d, and messaged %s\n", num, messg) ) ;
*/
void errprint(int err, char *fmt, ...) ;

/* use main log output/stdout settings , unconditional print */
void dbprint(char *fmt, ...) ;

/* output a hex buffer (uses mlogprint) */
void mlogprint_buffer(int dbg, unsigned char *buf, int first, int sz,
		      char *msg);

/* use main log output/stdout settings , conditional
   only if (lval & mainlog->debugval) : */
void mlogprint(int lval, char *fmt, ...) ;

#if 0
/* use update log output/stdout settings , conditional
   only if (lval & updlog->debugval)
   also send to main log file      */
void ulogprint(int lval, char *fmt, ...) ;
#endif

/* as mlogprint/ulogprint,  only prints a string */
void mlogpstr(int debval,char *str) ;
#if 0
void ulogpstr(int debval,char *str) ;
#endif

extern char time_stamp[32];
extern struct tm today_now;

void set_time_stamp() ;

#define mlogtimestamp(x) mlogpstr(x,time_stamp)
#if 0
#define ulogtimestamp(x) ulogpstr(x,time_stamp)
#endif


/* set tty_inputmode on/off (1/0) in unix,
   i.e. off: no echo and no canonical mode */
void tty_inputmode(int on) ;

#endif



