/* newcamd-client.c
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "confutil.h"
#include "config.h"
#include "common.h"
#include "des.h"

#include "newcamd-client.h"

#define CWS_NETMSGSIZE 240

typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint32;
typedef unsigned long long uint64;

#define CWS_FIRSTCMDNO 0xe0

extern char * __md5_crypt( const char *pw, const char *salt);

struct newcamd_server {
	/* These are data that are read from the configuration file */
	char *server_name;
	unsigned short port_number;
	char *password;
	char *username;
	unsigned char key[16];
	unsigned short ignorecws;
	
	/* These are data used at runtime */	
	int handle;
	unsigned char caid[2];
	unsigned char number_of_providers;
	unsigned char providers[16][3];
	unsigned char session_key[16];
	uint16 netMsgId;

	struct newcamd_server *next;
};

typedef enum
{
	MSG_CLIENT_2_SERVER_LOGIN = CWS_FIRSTCMDNO,
	MSG_CLIENT_2_SERVER_LOGIN_ACK,
	MSG_CLIENT_2_SERVER_LOGIN_NAK,
	MSG_CARD_DATA_REQ,
	MSG_CARD_DATA,
	MSG_SERVER_2_CLIENT_NAME,
	MSG_SERVER_2_CLIENT_NAME_ACK,
	MSG_SERVER_2_CLIENT_NAME_NAK,
	MSG_SERVER_2_CLIENT_LOGIN,
	MSG_SERVER_2_CLIENT_LOGIN_ACK,
	MSG_SERVER_2_CLIENT_LOGIN_NAK,
	MSG_ADMIN,
	MSG_ADMIN_ACK,
	MSG_ADMIN_LOGIN,
	MSG_ADMIN_LOGIN_ACK,
	MSG_ADMIN_LOGIN_NAK,
	MSG_ADMIN_COMMAND,
	MSG_ADMIN_COMMAND_ACK,
	MSG_ADMIN_COMMAND_NAK,
	MSG_KEEPALIVE = CWS_FIRSTCMDNO + 0x1d
} net_msg_type_t;

typedef enum
{
	COMMTYPE_CLIENT,
	COMMTYPE_SERVER
} comm_type_t;

static void des_key_parity_adjust(uint8 *key, uint8 len)
{
	uint8 i, j, parity;

	for (i = 0; i < len; i++)
	{
		parity = 1;
		for (j = 1; j < 8; j++) if ((key[i] >> j) & 0x1) parity = ~parity & 0x01;
		key[i] |= parity;
	}
}

static uint8 *des_key_spread(uint8 *normal)
{
	static uint8 spread[16];

	spread[ 0] = normal[ 0] & 0xfe;
	spread[ 1] = ((normal[ 0] << 7) | (normal[ 1] >> 1)) & 0xfe;
	spread[ 2] = ((normal[ 1] << 6) | (normal[ 2] >> 2)) & 0xfe;
	spread[ 3] = ((normal[ 2] << 5) | (normal[ 3] >> 3)) & 0xfe;
	spread[ 4] = ((normal[ 3] << 4) | (normal[ 4] >> 4)) & 0xfe;
	spread[ 5] = ((normal[ 4] << 3) | (normal[ 5] >> 5)) & 0xfe;
	spread[ 6] = ((normal[ 5] << 2) | (normal[ 6] >> 6)) & 0xfe;
	spread[ 7] = normal[ 6] << 1;
	spread[ 8] = normal[ 7] & 0xfe;
	spread[ 9] = ((normal[ 7] << 7) | (normal[ 8] >> 1)) & 0xfe;
	spread[10] = ((normal[ 8] << 6) | (normal[ 9] >> 2)) & 0xfe;
	spread[11] = ((normal[ 9] << 5) | (normal[10] >> 3)) & 0xfe;
	spread[12] = ((normal[10] << 4) | (normal[11] >> 4)) & 0xfe;
	spread[13] = ((normal[11] << 3) | (normal[12] >> 5)) & 0xfe;
	spread[14] = ((normal[12] << 2) | (normal[13] >> 6)) & 0xfe;
	spread[15] = normal[13] << 1;

	des_key_parity_adjust(spread, 16);
	return spread;
}

static void des_random_get(uint8 *buffer, uint8 len)
{
	uint8 idx = 0;
	int randomNo = 0;

	for (idx = 0; idx < len; idx++)
	{
		if (!(idx % 3)) randomNo = rand();
		buffer[idx] = (randomNo >> ((idx % 3) << 3)) & 0xff;
	}
}

static int des_encrypt(uint8 *buffer, int len, uint8 *deskey)
{
	uint8 checksum = 0;
	uint8 noPadBytes;
	uint8 padBytes[7];
	char ivec[8];
	uint16 i;

	if (!deskey) return len;
	noPadBytes = (8 - ((len - 1) % 8)) % 8;
	if (len + noPadBytes + 1 >= CWS_NETMSGSIZE-8) return -1;
	des_random_get(padBytes, noPadBytes);
	for (i = 0; i < noPadBytes; i++) buffer[len++] = padBytes[i];
	for (i = 2; i < len; i++) checksum ^= buffer[i];
	buffer[len++] = checksum;
	des_random_get((uint8 *)ivec, 8);
	memcpy(buffer+len, ivec, 8);
	for (i = 2; i < len; i += 8)
	{
		uint8 j;
		const uint8 flags = (1 << F_EURO_S2) | (1 << F_TRIPLE_DES);
		for(j=0; j<8; j++)
			buffer[i+j] ^= ivec[j];
                EuroDes(deskey, deskey+8, flags, 1, buffer+i);
		memcpy(ivec, buffer+i, 8);
	}
	len += 8;
	return len;
}

static int des_decrypt(uint8 *buffer, int len, uint8 *deskey)
{
	char ivec[8];
	char nextIvec[8];
	int i;
	uint8 checksum = 0;

	if (!deskey) return len;
	if ((len-2) % 8 || (len-2) < 16) return -1;
	len -= 8;
	memcpy(nextIvec, buffer+len, 8);
	for (i = 2; i < len; i += 8)
	{
		uint8 j;
                const uint8 flags = (1 << F_EURO_S2) | (1 << F_TRIPLE_DES);

		memcpy(ivec, nextIvec, 8);
		memcpy(nextIvec, buffer+i, 8);
                EuroDes(deskey, deskey+8, flags, 0, buffer+i);
      		for(j=0; j<8; j++)
                        buffer[i+j] ^= ivec[j];
	} 
	for (i = 2; i < len; i++) checksum ^= buffer[i];
	if (checksum) return -1;
	return len;
}

static uint8 *des_login_key_get(uint8 *key1, uint8 *key2, uint32 len)
{
	uint8 des14[14];
	static uint8 *des16;
	int i;

	memcpy(des14, key1, sizeof(des14));
	for (i = 0; i < len; i++) des14[i%14] ^= key2[i];
	des16 = des_key_spread(des14);
	doPC1(des16);
	doPC1(des16+8);
	return des16;
}

static int network_message_send(int handle, uint16 *netMsgId, uint8 *buffer, int len, uint8 *deskey, comm_type_t commType)
{
	uint8 netbuf[CWS_NETMSGSIZE];

	if (len < 3 || len + 12 > CWS_NETMSGSIZE || handle < 0) return -1;
	buffer[1] = (buffer[1] & 0xf0) | (((len - 3) >> 8) & 0x0f);
	buffer[2] = (len - 3) & 0xff;
	memcpy(netbuf+12, buffer, len);
	len += 12;
	if (netMsgId) { if (commType == COMMTYPE_CLIENT) (*netMsgId)++; netbuf[2] = (*netMsgId) >> 8; netbuf[3] = (*netMsgId) & 0xff; }
	else netbuf[2] = netbuf[3] = 0;
	memset(netbuf+4, 0, 8);
	if ((len = des_encrypt(netbuf, len, deskey)) < 0) return -1;
	netbuf[0] = (len - 2) >> 8;
	netbuf[1] = (len - 2) & 0xff;
	return write(handle, netbuf, len);
}

static int network_message_receive(int handle, uint16 *netMsgId, uint8 *buffer, uint8 *deskey, comm_type_t commType)
{
	int len;
	uint8 netbuf[CWS_NETMSGSIZE];
	int returnLen;

	if (!buffer || handle < 0) return -1;
	len = read(handle, netbuf, 2);
	if (!len) return 0;
	if (len != 2) return -1;
	if (((netbuf[0] << 8) | netbuf[1]) > CWS_NETMSGSIZE - 2) return -1;
	len = read(handle, netbuf+2, (netbuf[0] << 8) | netbuf[1]);
	if (!len) return 0;
	if (len != ((netbuf[0] << 8) | netbuf[1])) return -1;
	len += 2;
	if ((len = des_decrypt(netbuf, len, deskey)) < 15) return -1;
	if ((returnLen = (((netbuf[13] & 0x0f) << 8) | netbuf[14]) + 3) > len-12) return -1;
	if (netMsgId)
	{
		switch (commType)
		{
			case COMMTYPE_SERVER:
				*netMsgId = (netbuf[2] << 8) | netbuf[3];
				break;

			case COMMTYPE_CLIENT:
				if (*netMsgId != ((netbuf[2] << 8) | netbuf[3])) return -2;
				break;

			default:
				return -1;
				break;
		}
	}
	memcpy(buffer, netbuf+12, returnLen);
	return returnLen;
}

static void network_cmd_no_data_send(int handle, uint16 *netMsgId, net_msg_type_t cmd, uint8 *deskey, comm_type_t commType)
{
	uint8 buffer[CWS_NETMSGSIZE];

	buffer[0] = cmd; buffer[1] = 0;
	network_message_send(handle, netMsgId, buffer, 3, deskey, commType);
}

static int network_cmd_no_data_receive(int handle, uint16 *netMsgId, uint8 *deskey, comm_type_t commType)
{
	uint8 buffer[CWS_NETMSGSIZE];

	if (network_message_receive(handle, netMsgId, buffer, deskey, commType) != 3) return -1;
	return buffer[0];
}

int
connect_nonb(int sockfd, const struct sockaddr *saptr, socklen_t salen, int nsec)
{
	int				flags, n, error;
	socklen_t		len;
	fd_set			rset, wset;
	struct timeval	tval;

	flags = fcntl(sockfd, F_GETFL, 0);
	fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);

	error = 0;
	if ( (n = connect(sockfd, saptr, salen)) < 0)
		if (errno != EINPROGRESS)
			return(-1);

	/* Do whatever we want while the connect is taking place. */

	if (n == 0)
		goto done;	/* connect completed immediately */

	FD_ZERO(&rset);
	FD_SET(sockfd, &rset);
	wset = rset;
	tval.tv_sec = nsec;
	tval.tv_usec = 0;

	if ( (n = select(sockfd+1, &rset, &wset, NULL,
					 nsec ? &tval : NULL)) == 0) {
		close(sockfd);		/* timeout */
		errno = ETIMEDOUT;
		return(-1);
	}

	if (FD_ISSET(sockfd, &rset) || FD_ISSET(sockfd, &wset)) {
		len = sizeof(error);
		if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &error, &len) < 0)
			return(-1);			/* Solaris pending error */
	} else {
		return -2;
	}

done:
	fcntl(sockfd, F_SETFL, flags);	/* restore file status flags */

	if (error) {
		close(sockfd);		/* just in case */
		errno = error;
		return(-1);
	}
	return(0);
}

static int network_tcp_connection_open(uint8 *hostname, uint16 port)
{
        int handle;
        struct hostent *hostaddr;
        struct sockaddr_in socketAddr;
 
        if (!(hostaddr = gethostbyname(hostname))) { mlogprint(0xff, "Host lookup of %s failed\n", hostname); return -1; }
        if ((handle = socket(PF_INET, SOCK_STREAM, 0)) < 0) { mlogprint(0xff, "network make connection: couldn't create socket\n"); return -1; }
        socketAddr.sin_family = AF_INET;
        socketAddr.sin_port = htons(port);
        socketAddr.sin_addr.s_addr = ((struct in_addr *)hostaddr->h_addr)->s_addr;
        if (connect_nonb(handle, (struct sockaddr *)&socketAddr, sizeof(socketAddr), 5)
< 0) { mlogprint(0xff, "network make connection: error connect\n"); return -1; }
        return handle;
}

static int connect_newcamd_server(struct newcamd_server *server) {
	uint32 i;
	uint8 buf[CWS_NETMSGSIZE];
	uint8 keymod[14];
	uint8 *key;
  	int handle=0;

	uint32 index;
	uint8 *passwdcrypt;
	uint8 login_answer;
	int bytes_received;
	
	if(server->server_name == NULL || server->password == NULL || server->username == NULL ||
	   server->port_number == 0)
		return -5;

	handle = network_tcp_connection_open(server->server_name, server->port_number);
	if(handle < 0)
		return -1;
	server->handle = handle;
	server->netMsgId = 0;
	if(read(handle, keymod, sizeof(keymod)) != sizeof(keymod))
		return -2;

	key = des_login_key_get(keymod, server->key, 14);

	index = 3;
	buf[0] = MSG_CLIENT_2_SERVER_LOGIN;
	buf[1] = 0;
	strcpy(buf+index, server->username);
	passwdcrypt = (uint8*)__md5_crypt(server->password, "$1$abcdefgh$");
	index += strlen(server->username)+1;
	strcpy(buf+index, passwdcrypt);
	network_message_send(handle, &server->netMsgId, buf, index+strlen(passwdcrypt)+1, key, COMMTYPE_CLIENT);

	login_answer= network_cmd_no_data_receive(handle, NULL, key, COMMTYPE_CLIENT);
	if(login_answer != 0xe1) {
		return -3;
	}

	key = des_login_key_get(server->key, passwdcrypt, strlen(passwdcrypt));
	
	network_cmd_no_data_send(handle, &server->netMsgId, MSG_CARD_DATA_REQ, key, COMMTYPE_CLIENT);
	bytes_received = network_message_receive(handle, &server->netMsgId, buf, key, COMMTYPE_CLIENT);
	if(bytes_received < 16 || buf[0] != MSG_CARD_DATA) {
		return -4;
	}

	server->caid[0] = buf[4];
	server->caid[1] = buf[5];
	server->number_of_providers = buf[14];
	
	for(i=0; i<server->number_of_providers; i++) {
		server->providers[i][0] = buf[15+11*i];
                server->providers[i][1] = buf[16+11*i];
                server->providers[i][2] = buf[17+11*i];

	}

	memcpy(server->session_key, key, 16);
	return 0;
}

static struct newcamd_server *servers = NULL;

static int ConfigVarVal(char *var, char *val, short system)
{
  extern unsigned short rcaid;                       //radegast caid; musi byt "unsigned short", "int"... 
  int res=1;
	if(!StrNoCaseCmp(var,"port")) {
		unsigned long port;
		if ((res=SetUL(val,&port,2,10))) {
			servers->port_number = (unsigned short)port;
		}
	} else if(!StrNoCaseCmp(var,"caid")) {
		unsigned long caid;
    
		if ((res=SetUL(val,&caid,2,16))) {
			servers->caid[0] = (caid>>8) & 0xff;
			servers->caid[1] = caid & 0xff;
      rcaid=caid;
      mlogprint(0xff, "Radegast CAID %04x \n", rcaid);
		}
	} else if(!StrNoCaseCmp(var,"username")) {
		servers->username = (char*)malloc(strlen(val)+1);
		strcpy(servers->username, val);
	} else if(!StrNoCaseCmp(var,"password")) {
                servers->password = (char*)malloc(strlen(val)+1);
                strcpy(servers->password, val);
        } else if(!StrNoCaseCmp(var,"key")) {
    		res=SetU8(val,servers->key,14,16,0,255);
	} else if(!StrNoCaseCmp(var,"ignore")) {
                servers->ignorecws = 1;
	} else {
		res = 0;
	}

	return !res;
}

static int ConfigSection(char *s, short system) {
	struct newcamd_server *new_server = (struct newcamd_server *)malloc(sizeof(struct newcamd_server));
	memset(new_server, 0, sizeof(struct newcamd_server));
	new_server->next = servers;
	servers = new_server;

	servers->server_name = (char *)malloc(strlen(s)+1);
	strcpy(servers->server_name, s);
	return 0;
}

static int connect_radegast_server(struct newcamd_server *server) {
  	int handle;

	if(server->server_name == NULL || server->port_number == 0)
		return -5;
	
	handle= network_tcp_connection_open(server->server_name, server->port_number);
	if(handle < 0)
		return -1;
	server->handle = handle;

	return 0;
}

static int server_connect(struct newcamd_server *server_it) {
	int res;
	if(server_it->username != NULL &&
	   server_it->password != NULL) {
		res = connect_newcamd_server(server_it);
	} else {
		res = connect_radegast_server(server_it);
	}
	return res;
}
static int radegast_send_ecm(int handle, int sysid, int provid, int keynr, unsigned char *ecm, int length) {
  
 	unsigned char outmsg[512];
	int ix = 0;

	ix = 0;
        memset(outmsg, '0', 256);
	
        outmsg[ix++] = '\1';    /* CMD_ECM_KEY_ASK */
        outmsg[ix++] = 0;       /* len */
	      outmsg[ix++] = 10;       /* ECM_SYSTEM_ID */
        outmsg[ix++] = 2;       /* len */
        outmsg[ix++] = sysid >> 8;
        outmsg[ix++] = sysid & 0xff;
	      outmsg[ix++] = 6;       /* ECM_PROVIDER_ID */
        outmsg[ix++] = 8;       /* len */
        sprintf(&outmsg[ix], "%08X", provid);
        
        ix += 8;
	      outmsg[ix++] = 7;       /* ECM_KEY_NUMBER */
        outmsg[ix++] = 4;       /* len */
        sprintf(&outmsg[ix], "%04X", keynr);
        
        ix += 4;
        outmsg[ix++] = 3;       /* ECM_CA_DATA */
        outmsg[ix++] = length & 0xff;     /* len */
        memcpy(&outmsg[ix], ecm, length);
        
        ix += length;
        outmsg[1] = ix-2;       /* len */
        
        //printf("kontrola len %x.\n", length);
                
	return write(handle, outmsg, ix);
}

static int radegast_get_cw(int handle, char *plainwords) {
	unsigned char inbuf[256];
	int rv;

	rv = read(handle, inbuf, 2);
	if(rv != 2) {
		return -5;
	}
	rv = read(handle, inbuf+2, inbuf[1]);
	if(rv != inbuf[1]) {
		return -2;
	}

	if(inbuf[0] != 2) {
		return -3;
	}
	if(inbuf[2] != 4 && inbuf[2] != 5) {
		return -4;
	}
	if(inbuf[2] == 4) {
		/* Wrong key */
		return 4;
	} else {
		/* Key working */
		memcpy(plainwords, &inbuf[4], 16);
		return 20;
	}
	
	return 0;
}

int newcamd_decode_ecm(unsigned short caid, unsigned char *ecm, int length, unsigned char *plainwords, newcamd_callback callback) {
	int res = -3;
	struct newcamd_server *server_it, *prev = NULL;
	
	for(server_it = servers; server_it; prev = server_it, server_it = server_it->next) {
	/*  if((caid >> 8) == server_it->caid[0] &&
		   (caid & 0xff) == server_it->caid[1]) {  */ 
			uint8 buf[CWS_NETMSGSIZE]; 
			if(server_it->handle <= 0) {
				int res = server_connect(server_it);
				if(res < 0) {
					if(server_it->handle > 0) {
						close(server_it->handle);
						server_it->handle = 0;
					}
					continue;
				}
			}

			if(server_it->username != NULL &&
			   server_it->password != NULL) {
				int bytes_received = 3;
				int retries = 8;
				/* Silly way to detect if we are talking to a newcamd server */
				if(network_message_send(server_it->handle, &server_it->netMsgId, ecm, length, server_it->session_key, COMMTYPE_CLIENT) > 0) {
					while(retries--) {
						fd_set rfds;
						struct timeval tv;
						
						FD_ZERO(&rfds);
						FD_SET(server_it->handle, &rfds);
						tv.tv_sec = 0;
						tv.tv_usec = 350*1000;
						if(select(server_it->handle+1, &rfds, NULL, &rfds, &tv) == 0) {
							if(callback != NULL)
								callback();
						} else {
							bytes_received = network_message_receive(server_it->handle, &server_it->netMsgId, buf, server_it->session_key, COMMTYPE_CLIENT);
							if(bytes_received == -2) {
								/* Reply from an old packet */
								continue;
							}
							break;
						}					
					}
				} else {
					bytes_received = -1;
				}
				if(bytes_received < 0) {
					mlogprint(0x10, "\nServer %s went down\n", server_it->server_name);
					/* Connection must be down */
					close(server_it->handle);
					server_it->handle = 0;
					res = -2;
				} else if(bytes_received == 3) {
					mlogprint(0x10, "\n[%04x] Server %s cant decrypt\n", caid, server_it->server_name);
					/* Card wasn't able to decrypt channel */
					res = -1;
				} else if(bytes_received == 19) {
					mlogprint(0x10, "\nServer %s replied ok - port %04d\n", server_it->server_name,server_it->port_number);
					memcpy(plainwords, buf+3, 16);
					res = 0;
					break;
				} else {
					/* Wow - this was unexpected */
					res = -4;
				}
			} else {
				int bytes_received = 4;
				int retries = 8;
				//if(radegast_send_ecm(server_it->handle, (caid >> 8) & 0xff, 0, 0, ecm, length) > 0)
        if(radegast_send_ecm(server_it->handle, caid, 0, 0, ecm, length) > 0)
         {
         	/* Radegast server it is */
					while(retries--) {
						fd_set rfds;
						struct timeval tv;
						
						FD_ZERO(&rfds);
						FD_SET(server_it->handle, &rfds);
						tv.tv_sec = 0;
						tv.tv_usec = 350*1000;
						if(select(server_it->handle+1, &rfds, NULL, &rfds, &tv) == 0) {
							if(callback != NULL)
							callback();
						} else {
							bytes_received = radegast_get_cw(server_it->handle, plainwords);
							break;
						}					
					}					
				} else {
					bytes_received = -1;
				}
				if(bytes_received < 0) {
					/* Connection must be down or giving us trouble */
					close(server_it->handle);
					server_it->handle = 0;
					res = -2;
				} else if(bytes_received == 4) {
					/* Card wasn't able to decrypt channel */
					res = -1;
				} else if(bytes_received == 20) {
					res = 0;
					break;
				} else {
					/* We should not be here */
					res = -1;
				}
			}
	//	}
	}

	if(res == 0) {
		if(prev != NULL) {
			prev->next = server_it->next;
			server_it->next = servers;
			servers = server_it;
		}
	}

	return res;
}

int init_newcamd() {
	struct newcamd_server *server_it;

	if(ReadConfigFile("newcamd-client.dat", &ConfigSection, &ConfigVarVal, 0) != 0)
		return -3;

	/* Now connect to the servers */
	for(server_it = servers; server_it; server_it = server_it->next) {
	  if (server_it->ignorecws != 1) {
		int res = server_connect(server_it);
		if(res < 0) {
			mlogprint(0xff, "Connection to %s failed with error %d\n", server_it->server_name, res);
			if(server_it->handle > 0) {
				close(server_it->handle);
				server_it->handle = 0;
			}
		} else {
      mlogprint(0xff, "Got CA ID %02x%02x from %s - port %04d\n", server_it->caid[0], server_it->caid[1], server_it->server_name,server_it->port_number);
		}
	  }
	}
	return 0;
}
