/*  serfunc.c
      some demo functions to manipulate serial port and speed in linux
      probably kernel >= 2.2.x is needed
    Copyright (C) 2000  The Linux Card Emulator Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    http://www.fsf.org/
*/

/*
  sendbyte() types: (for log)

  - 1 byte (ack)
  - swok/err 90 00 / 90 08 / 90 40

  sendbyte_n(dbgval, *char , n)
  - read "n" bytes and put in log


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef USE_PRIORITY_SCHEDULING
#include <sched.h>
#endif

/* #define DBOX2 1 */

/* for speeds etc. look at following: */
#if !defined(__CYGWIN__) && !defined(DBOX2)
#include <linux/serial.h>
#endif

#include <termios.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>

#include "debug.h"
#include "serfunc.h"
#include "main.h"
#include "config.h"


#define OLDKERNEL
#define HAVE_SELECT 1
#define STDSERIAL 0
#define HIGHSPEED 0

#ifndef ASYNC_INTERNAL_FLAGS
#define ASYNC_INTERNAL_FLAGS 0xFF000000
#endif
#define SPD_CUST ASYNC_INTERNAL_FLAGS|ASYNC_SPD_CUST
#define SPEED_NOT_FOUND ASYNC_INTERNAL_FLAGS

#define CHKRESET 1

typedef struct lserial_port_info {
	int fd;	  /* -1 if not initialized */
        char *devname;
	struct termios termios_data;
#if !defined(__CYGWIN__) && !defined(DBOX2)
	struct serial_struct sst;
#endif
#if defined(DBOX2)
	int rtsctsfd;
#endif
	int reset;
	speed_t speed;
        int bytereverse;
        tcflag_t serparity, cstopb, cbitsize;
        int flags;
	int reset_detect_mask;
	int phoenix;
} serial_port_info;


/* some local declarations : */
int open_serial_port_season(int deb, char *sname) ;
int open_serial_port_phoenix(int deb, char *sname) ;
void set_bps(int portno, int bps) ;
static int readbyte_port(int deb, long useconds, int flags ) ;
int readbyte_norm(serial_port_info *port, long useconds, int flags) ;
/* print port info from  to file f */
int port_fprintf_info(FILE *f, int deb) ;
static int check_and_handle_reset(int flags) ;

/* delay byte2del=BYTE2D is in mikroseconds ,
   for 9600-10000bps 8bit+ parity bytes are received in approx. 1000 bytes/sec
   => minimum 0.9 millisecons/byte *2 = 1.8 milliseconds/byte
   select() returns when data has been received, so the value should
   at least 2 millisecons
*/
/* delay reset2d=RESET2D is in mikroseconds

   According to ISO 7816:
   After Reset goes Low (active):
     Answer to reset should be sent earliest 400, latest in 40000 clock cycles
     @10000bps or 3.72MHz :  min 100us ... max 10ms
     and should be sent while RESET is avtive
   If no answer to reset (ATR) during Reset=Low,
   ATR should be sent earliest 400,  and latest in 40000 clock cycles
   after Reset transition from Low to High (inactive).
 */
#if 0
 #define RESETD 19
 #define RESET2D 0
 /* read byte max wait for sendbyte [usec] */
 #define BYTED 2500
 /* read byte max wait for byte [usec]*/
 #define BYTE2D 1400

 #define RESETD 12
 #define RESET2D 0
 #define BYTED 10000
 #define BYTE2D 24000
#endif

static int have_reset=0;

static char *devname=NULL, devttySn[]="/dev/ttyS0";

static int micro_sleep(int n) {
  /* call POSIX nanosleep(n*1000) : */
#ifdef __CYGWIN__
  return usleep(n);
#else
  static struct timespec tmspec = { 0L, 1000000L };
  unsigned long nl= n;
  tmspec.tv_nsec= (nl%1000000L)*1000L+1;
  tmspec.tv_sec= 0L; /* nl/1000000UL; */
  /* printf(" {%d => %lu.%09lu} ",n,tmspec.tv_sec, tmspec.tv_nsec); */
  return nanosleep(&tmspec, NULL);
#endif
}

static serial_port_info my_port[2];
static serial_port_info *port= my_port ;

typedef struct speed_bps {
#ifdef __CYGWIN__
        int speed;
#else
	speed_t speed;
#endif
	int bps;
	char *s;
} speed_bps_t ;

static speed_bps_t  speedbps_tbl[]={
   { B0,  0 , "B0"},
   { B50,  50, "B50" },
   { B75,  75 , "B75"},
   { B110,  110, "B110" },
   { B134,  134, "B134" },
   { B150,  150 ,"B150"},
   { B200,  200 , "B200"},
   { B300,  300 , "B300"},
   { B600,  600 , "B600"},
   { B1200,  1200 , "B1200"},
   { B1800,  1800 , "B1800"},
   { B4800,  4800 , "B4800"},
   { B9600,  9600 , "B9600"},
   { B19200,  19200 , "B19200"},
   { B38400,  38400 , "B38400"},
   { B57600,  57600 , "B57600"},
   { B115200,  115200 , "B115200"},
#ifndef __CYGWIN__
   { B230400,  230400 , "B230400"},
   { B460800,  460800 , "B460800"},
#ifndef OLDKERNEL
   { B500000,   500000 ,  "B500000"},
   { B576000,   576000 ,  "B576000"},
   { B921600,   921600 ,  "B921600"},
   { B1000000,  1000000, "B1000000"},
   { B1152000,  1152000, "B1152000"},
   { B1500000,  1500000, "B1500000" },
   { B2000000,  2000000, "B2000000" },
   { B2500000,  2500000, "B2500000" },
   { B3000000,  3000000, "B3000000" },
   { B3500000,  3500000, "B3500000" },
   { B4000000,  4000000, "B4000000" },
#endif
#endif
   { SPEED_NOT_FOUND,  -1 , "unknown"}
};

static speed_bps_t * speed_to_bps(speed_t sp) {
  speed_bps_t *tsp= &speedbps_tbl[0];
  while (tsp->bps>=0 && tsp->speed !=sp) tsp++;
  return tsp;
}

#if !defined(__CYGWIN__) && !defined(DBOX2)
static speed_bps_t * bps_to_speed(int bps) {
  speed_bps_t *tsp= &speedbps_tbl[0];
  while (tsp->bps>=0 && tsp->bps !=bps) tsp++;
  return tsp;
}
#endif

void do_port_error(char *s) {
  errprint(errno,s);
  errprint(0,"Verify that %s is a valid and operational tty port.\n",devname);
  /* old:  exit(1); */
}

typedef struct str_int {
	char *s;
	int  i;
} str_int_t ;

#if !defined(__CYGWIN__) && !defined(DBOX2)
static str_int_t stype_tbl[]= {
	{"unknown",PORT_UNKNOWN},
	{"8250",PORT_8250},
	{"16450",PORT_16450},
	{"16550",PORT_16550},
	{"16550A",PORT_16550A},
	{"CIRRUS",PORT_CIRRUS},
	{"16650",PORT_16650},
	{"16650V2",PORT_16650V2},
	{"16750",PORT_16750},
	{"STARTECH",PORT_STARTECH},
	{"**",-1}
};

static char * serialporttype(int tp) {
   str_int_t *t= &stype_tbl[0];
   while ( t->i>=0 && t->i!=tp) t++;
   if (t->i<0) t= &stype_tbl[0];
   return t->s;
}
#endif

#if 0
/* only possible as root ? */
/*
static void set_baud_base(int deb,int baudbase) {
  int r;
  if (port->fd<0) return ;
  r=ioctl(port->fd, TIOCGSERIAL, &port->sst );
  xxxx port->sst.type=PORT_8250;
  port->sst.baud_base= baudbase;
  xxxx port->sst.xmit_fifo_size=1;
  r=ioctl(port->fd, TIOCSSERIAL, &port->sst );
}
*/
#endif

enum { RESET_IGNORE=0x0, RESET_DETECT=0x1 };
void set_reset_detection(int det);

void set_reset_detection(int det) {
  if (det==RESET_IGNORE) port->flags &= ~RESET_DETECT;
  else if (det==RESET_DETECT) port->flags |= RESET_DETECT;
}

int get_bps_i(int io) {
  /* io= 0 output / 1 input */
  int r,bps=9600;
  if (port->fd<0) return 0;

  r= tcgetattr ( port->fd, &port->termios_data );
#if !defined(__CYGWIN__) && !defined(DBOX2)
  r=ioctl(port->fd, TIOCGSERIAL, &port->sst );
  r=ioctl(port->fd,  TIOCMBIS, TIOCM_DTR | TIOCM_RTS , &port->sst );
  port->speed= io? cfgetospeed ( &port->termios_data ) : cfgetispeed ( &port->termios_data );
  /*printf("\t%cspeed= 0x%x\n",io?'i':'o',port->speed); */
  if ( (port->speed==B38400 && (port->sst.flags&ASYNC_SPD_CUST)) && port->sst.custom_divisor) {
	bps= port->sst.baud_base/port->sst.custom_divisor;
  } else bps= speed_to_bps(port->speed)->bps;
#endif
  return bps;
}


int open_serial_port_season(int deb, char *sname) {
  int r;

  if (port->devname) free(port->devname);
  if (sname) {
    port->devname = calloc(strlen(sname)+1,sizeof(char));
    strcpy(port->devname,sname);
  } else {
    port->devname=NULL;
    return -1;
  }

  port->reset=0;
  port->reset_detect_mask= TIOCM_CAR;

  SERDEBUGPR(0xfe,("Open 0) [%s]\n",port->devname));

  port->fd= open(port->devname,O_RDWR |  O_NDELAY /* |O_NONBLOCK */);
  if (port->fd<0) errprint(errno,"open port\n");

#ifdef DBOX2
  port->rtsctsfd = open("/dev/rtscts", O_RDWR);
  if (port->rtsctsfd<0) errprint(errno,"open port (you need to load camd-cardserver.o)\n");
#endif

  SERDEBUGPR(0xfe,("opened %s\n",port->devname));
  if (port->fd>=0) {
    r= tcgetattr ( port->fd, &port->termios_data );
#ifndef __CYGWIN__
#ifndef DBOX2
    r=ioctl(port->fd, TIOCGSERIAL, &port->sst );
#endif
#if 0
    SERDEBUGPR(0xfe,("open 2)\n"));
    port->termios_data.c_iflag |= BRKINT| INPCK;
    port->termios_data.c_iflag &= ~(IXON | IXOFF | IXANY);
#endif
#if STDSERIAL
    port->termios_data.c_cflag =  CS8  | CRTSCTS  |CLOCAL | HUPCL ;
#else
    SERDEBUGPR(0xa,(" cdata=0x%x ",port->termios_data.c_cflag));

    /* cfmakeraw sets the terminal attributes as follows:
                   termios_p->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP
                                   |INLCR|IGNCR|ICRNL|IXON);
                   termios_p->c_oflag &= ~OPOST;
                   termios_p->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
                   termios_p->c_cflag &= ~(CSIZE|PARENB);
                   termios_p->c_cflag |= CS8;  */
    cfmakeraw (&port->termios_data );

    port->termios_data.c_cflag=B9600;
    SERDEBUGPR(0xa,("cdata=0x%x ",port->termios_data.c_cflag)) ;
    /* enable receiver, N bits and 1-2 stop bit : */
    port->termios_data.c_cflag |= CREAD | (CSIZE & port->cbitsize ) | port->cstopb ;
    /* first set parity bits to 0 , then set from serparity: */
    port->termios_data.c_cflag &=  ~(PARENB | PARODD);
    port->termios_data.c_cflag |= ((PARENB | PARODD) & port->serparity) ;

    /* ignore modem control l ines   */
    port->termios_data.c_cflag |= CLOCAL;
    /* lower modem control lines after last process closes
              the device (hang up). */
    port->termios_data.c_cflag |= HUPCL ;
    /* flow control. */
    /* port->termios_data.c_cflag |= CRTSCTS; */
    SERDEBUGPR(0xa,("cdata=0x%x ",port->termios_data.c_cflag));
#endif
#endif
    /* port->termios_data.c_oflag = port->termios_data.c_lflag = 0;
       r= tcsetattr ( port->fd, TCSANOW, &port->termios_data );*/
    SERDEBUGPR(0xa,("cdata=0x%x\n",port->termios_data.c_cflag));
    SERDEBUGPR(0xa,("c_oflag=0x%x c_iflag=0x%x ->",port->termios_data.c_oflag,
	   port->termios_data.c_iflag));
    port->termios_data.c_iflag=BRKINT|IXANY;
    port->termios_data.c_oflag=0;
    SERDEBUGPR(0xa,("c_oflag=0x%x c_iflag=0x%x\n",port->termios_data.c_oflag,
	   port->termios_data.c_iflag));
    r= tcsetattr ( port->fd, TCSAFLUSH, &port->termios_data );
   sleep(1);
#if 1
    port->fd = open(port->devname,O_RDWR); /* Reopen line with CLOCAL */
    if (port->fd<=0) errprint(errno,"reopen port \n");
#endif
  }
  return port->fd;
}

int open_serial_port_phoenix(int deb, char *sname) {
  int r;

  if (port->devname) free(port->devname);
  if (sname) {
    port->devname = calloc(strlen(sname)+1,sizeof(char));
    strcpy(port->devname,sname);
  } else {
    port->devname=NULL;
    return -1;
  }

  port->reset=0;
  port->reset_detect_mask= TIOCM_CTS;

  SERDEBUGPR(0xfe,("Open 0) [%s]\n",port->devname));

  port->fd= open(port->devname,O_RDWR |  O_NDELAY /* |O_NONBLOCK */);
  if (port->fd<0) errprint(errno,"open port\n");

  SERDEBUGPR(0xfe,("opened %s\n",port->devname));
  if (port->fd>=0) {
    r= tcgetattr ( port->fd, &port->termios_data );
#ifndef __CYGWIN__
#ifndef DBOX2
    r=ioctl(port->fd, TIOCGSERIAL, &port->sst );
#endif
#if 0
    SERDEBUGPR(0xfe,("open 2)\n"));
    port->termios_data.c_iflag |= BRKINT| INPCK;
    port->termios_data.c_iflag &= ~(IXON | IXOFF | IXANY);
#endif
#if STDSERIAL
    port->termios_data.c_cflag =  CS8  | CRTSDTR  |CLOCAL | HUPCL ;
#else
    SERDEBUGPR(0xa,(" cdata=0x%x ",port->termios_data.c_cflag));

    /* cfmakeraw sets the terminal attributes as follows:
                   termios_p->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP
                                   |INLCR|IGNCR|ICRNL|IXON);
                   termios_p->c_oflag &= ~OPOST;
                   termios_p->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
                   termios_p->c_cflag &= ~(CSIZE|PARENB);
                   termios_p->c_cflag |= CS8;  */
    cfmakeraw (&port->termios_data );

    port->termios_data.c_cflag=B9600;
    SERDEBUGPR(0xa,("cdata=0x%x ",port->termios_data.c_cflag)) ;
    /* enable receiver, N bits and 1-2 stop bit : */
    port->termios_data.c_cflag |= CREAD | (CSIZE & port->cbitsize ) | port->cstopb ;
    /* first set parity bits to 0 , then set from serparity: */
    port->termios_data.c_cflag &=  ~(PARENB | PARODD);
    port->termios_data.c_cflag |= ((PARENB | PARODD) & port->serparity) ;

    /* ignore modem control l ines   */
    port->termios_data.c_cflag |= CLOCAL;
    /* lower modem control lines after last process closes
              The device (hang up). */
    port->termios_data.c_cflag |= HUPCL ;
    /* flow control. */
    /* port->termios_data.c_cflag |= CRTSCTS; */
    SERDEBUGPR(0xa,("cdata=0x%x ",port->termios_data.c_cflag));
#endif
#endif
    /* port->termios_data.c_oflag = port->termios_data.c_lflag = 0;
       r= tcsetattr ( port->fd, TCSANOW, &port->termios_data );*/
    SERDEBUGPR(0xa,("cdata=0x%x\n",port->termios_data.c_cflag));
    SERDEBUGPR(0xa,("c_oflag=0x%x c_iflag=0x%x ->",port->termios_data.c_oflag,
	   port->termios_data.c_iflag));
    port->termios_data.c_iflag=BRKINT|IXANY;
    port->termios_data.c_oflag=0;
    SERDEBUGPR(0xa,("c_oflag=0x%x c_iflag=0x%x\n",port->termios_data.c_oflag,
	   port->termios_data.c_iflag));
    r= tcsetattr ( port->fd, TCSAFLUSH, &port->termios_data );
   sleep(1);
#if 1
    port->fd = open(port->devname,O_RDWR); /* Reopen line with CLOCAL */
    if (port->fd<=0) errprint(errno,"reopen port \n");
#endif
  }
  return port->fd;
}

#define SETBAUDRATE_0	_IOW('o', 7, unsigned int)
#define GET_PCDAT	_IOR('o', 6, unsigned int)

void set_bps(int portno, int bps) {
#if !defined(__CYGWIN__) && !defined(DBOX2)
  int r;
  speed_bps_t *sbps;

  if (port->fd<0) return;

  r=ioctl(port->fd, TIOCGSERIAL, &port->sst );

  if (bps<=0) {
    /* set_bps() called with non-positive BPS value, makes no sense,
       so give port->speed some sane value like B9600 */
    port->speed= B9600;
    SERDEBUGPR(0xffff,("error! set_bps(%d,%d) setting to B9600 [0x%0x]\n",portno,bps,port->speed));
  } else {
    speed_bps_t *sbps_o;
    sbps_o = bps_to_speed(bps);
    port->speed= sbps_o->speed;
    if (port->speed==B38400) {
      /* The B38400 setting may be used for custom settings with
	 ASYNC_SPD_CUST , so reset custom_divisor so that 38400 is used */
      port->sst.custom_divisor= 0;
      port->sst.flags &= ~ASYNC_SPD_CUST;
      SERDEBUGPR(0xfe,("set_bps(,%d) %s [0x%0x]\n",bps,sbps_o->s,port->speed));
    } else if (port->speed == SPEED_NOT_FOUND  ) {
      /* not found bps in table, so use internal custom speed
	 and calculate nearest possible bps rate with custom_divisor */
      port->sst.custom_divisor= (port->sst.baud_base*10/bps+4)/10;
      port->speed= B38400 ;
      port->sst.flags |= ASYNC_SPD_CUST; /* set spd_cust */
      SERDEBUGPR(0xfe,("set_bps(,%d) ASYNC_SPD_CUST (B38400) %s [0x%0x] divisor=%d\n",
		       bps,sbps_o->s,port->speed,port->sst.custom_divisor));
    } else
      SERDEBUGPR(0xfe,("set_bps(,%d) std %s [0x%0x]\n",bps,sbps_o->s,port->speed));
    /*
	port->speed already contains standard value from the table speed_bps_t
    */
  }
  /* port->sst.xmit_fifo_size=1; */
  r=ioctl(port->fd, TIOCSSERIAL, &port->sst );
  r= tcgetattr ( port->fd, &port->termios_data );
  r=cfsetospeed(&port->termios_data,port->speed);
  r=cfsetispeed(&port->termios_data,port->speed);
  sbps= speed_to_bps(port->speed);
  SERDEBUGPR(0xa,("\tset speed= 0x%x %s (wanted %d bps)\n", port->speed,sbps->s, bps));
  r= tcsetattr ( port->fd, TCSAFLUSH, &port->termios_data );
  if (r) SERDEBUGPR(0xa,(" ioctrl:%d\n",r));
#endif

#ifdef DBOX2
  if(port->rtsctsfd > 0) {
    const int CPUSPEED = 67000000;
    int brgc = ((CPUSPEED >> 4) / bps) - 1;
    int mod = (CPUSPEED >> 4) % bps;
    if (mod > (bps >> 1)) brgc++;

    ioctl(port->rtsctsfd, SETBAUDRATE_0, brgc);
  }
#endif
}


int port_fprintf_info(FILE *f,int deb) {
  int r;
  speed_t ospeed,ispeed;
  speed_bps_t *ispd,*ospd;

  if (port->fd<0) return -1;

   r= tcgetattr ( port->fd, &port->termios_data );
   if (r<0) {
       do_port_error("tcgetattr() open port error\n");
       return -1;
   }
   ospeed=cfgetospeed ( &port->termios_data );
   ispeed=cfgetispeed ( &port->termios_data );
#if !defined(__CYGWIN__) && !defined(DBOX2)
   r=ioctl(port->fd, TIOCGSERIAL, &port->sst );
   if (r<0) {
       do_port_error("ioctl() error\n");
       return -2;
   }
   SERDEBUGPR(0xfe,("\ttype=%d (%s) ",port->sst.type,serialporttype(port->sst.type)));
   SERDEBUGPR(0xfe,("line=%d port=0x%x irq=0x%x\n",port->sst.line,
		port->sst.port,port->sst.irq));
   SERDEBUGPR(0xfe,("\tflags==0x%x\n",port->sst.flags));
   SERDEBUGPR(0xfe,("\txmit_fifo_size=%d\n",port->sst.xmit_fifo_size));
   SERDEBUGPR(0xfe,("\tcustom_divisor=%d\n",port->sst.custom_divisor));
   SERDEBUGPR(0xfe,("\tbaud_base=%d\n",port->sst.baud_base));
#endif

   ispd=speed_to_bps(ispeed);
   ospd=speed_to_bps(ospeed);
   SERDEBUGPR(0xfe,("\tispeed=%d (%s / %d),  ospeed=%d (%s / %d)\n",
	ispeed,ispd->s, get_bps_i(1),
	ospeed,ospd->s, get_bps_i(0)));
   return 0;
}

void serial_port_shutdown(int deb) {
  /* should restore serial ports ? */
}

static int bits4[16] ;

static int do_bits4(int b) {
  b &= 0xf;
  return ((b>>3)&0x1)
	|((b>>1)&0x2)
	|((b<<1)&0x4)
	|((b<<3)&0x8) ;
}

static void init_bits4(){
  int i;
  for (i=0;i<16; i++) bits4[i]= do_bits4(i);
};

/* handle debug/log different if "logcard" flag i set */
static int comm_opt_logcard =0;
void set_logcard(int logcard) {  comm_opt_logcard= logcard; }

static int swapbits(int b) {
  return 0xff & ( port->bytereverse ?
		    ~(((bits4[b&0xf]<<4)&0xf0)|(bits4[(b>>4)&0xf])) : b);
}

int  sendbyte(int deb,int ch) {
  /* return :
     < 0 : error
     = 0 : ok
  */
  int rdc,status=0;
  char c ;

  if (port->fd<0) return -1;
  if (check_and_handle_reset(CHKRESET | SERDB_WR)) return -1;
  c= swapbits(ch);

  if (!comm_opt_logcard) {
    if(write(port->fd, &c, 1)<0){
      SERDEBUGPR(SERDB_WR,("{%02x,%02x} ",ch&0xff,c));
      errprint(errno,"Verify that %s is a valid and operational tty port.\n",port->devname);
    } else {
      SERDEBUGPR(SERDB_WR,("[w:%02x]",ch&0xff));
      if (deb) mlogprint(deb," %02x", ch&0xff);
    }
    fsync(port->fd);
  }
  /*micro_sleep(bytedel); */
  if ((rdc=readbyte_port(deb,serial_delay.curr_wr,CHKRESET | SERDB_WR))<0) {
    status= -1;
    if (status<0) SERDEBUGPR(0xffff,(" rcv.err:%d ",rdc));
  } else {
    if (comm_opt_logcard) {
      mlogprint(deb," %02x", (rdc&0xff));
      status=0;
    } else {
      status= (rdc==ch)? rdc: -1;
      /*micro_sleep(1000); */
      if (status<0) SERDEBUGPR(0xffff,(" rcv.err[%02x!=%02x]",ch,rdc));
    }
  }
  return status;
}

enum { SerBufsz=12 };

int  sendbyte_n(int deb,byte *chs, int n) {
  /* return number of written bytes
     or < 0 if error
  */
  static char wbuf[SerBufsz];
  static byte *lchs;
  int rem,cnt,rdc,status=0;
  char *cp;

  if (port->fd<0) return -1;
  if (check_and_handle_reset(CHKRESET | SERDB_WR)) return -1;
  if (n<=0) return -1;
  rem=n;
  /*SERDEBUGPR(0x6,("(%d):: ",n));*/
  while (rem>0) {
    int bufn;
    if (comm_opt_logcard) {
      bufn=rem;
    } else {
      bufn= rem>SerBufsz? SerBufsz: rem;
      cp= wbuf;
      lchs=chs;
      for (cnt=0 ; cnt<bufn; cnt++)  {
	byte cc1,cc2;
	cc1= *(lchs++);
	cc2= swapbits(cc1);
	SERDEBUGPR(0x6,(" %02x->%02x ",cc1,cc2));
	*(cp++)=cc2;
      }
      /*SERDEBUGPR(0x6,(" (%d)\n",bufn));*/
      rem -= bufn;
      if (write(port->fd, wbuf, bufn)<0) {
	cp= wbuf;
	if (serdebug&SERDB_WR) {
	  int i;
	  SERDEBUGPR(SERDB_WR,(" w["));
	  for (i=0; i<bufn; i++)
	    SERDEBUGPR(SERDB_WR,(" %02x",*(cp++)&0xff));
	  SERDEBUGPR(SERDB_WR,("]\n"));
	}
	errprint(errno,"Verify that %s is a valid and operational tty port.\n",port->devname);
	break;
      } else {
	int i;
	cp=wbuf;
	if (serdebug&SERDB_WR) {
	  SERDEBUGPR(SERDB_WR,("w["));
	  for (i=0; i<bufn; i++)
	    SERDEBUGPR(SERDB_WR,(" %02x",*(cp++)&0xff));
	  SERDEBUGPR(SERDB_WR,("] "));
	}
	lchs=chs;
	if (deb) for (i=0; i<bufn; i++)
	  mlogprint(deb," %02x", *(lchs++)&0xff);
	flush_win();
      }
    }
    for ( ; bufn>0; bufn--, chs++) {
      rdc=readbyte_port(deb,serial_delay.curr_wr,CHKRESET | SERDB_WR);
      if (rdc<0) {
	status= -1;
	SERDEBUGPR(0xffff,(" rcv.err:%d ",rdc));
	return status;
      } else {
	if (comm_opt_logcard) {
	  status=0;
	  mlogprint(deb," %02x", (rdc&0xff));
	} else {
	  status= ((rdc&0xff)== *chs)? rdc: -1;
	  if (status<0 )
	    SERDEBUGPR(0xffff,(" rcv.err[%02x!=%02x]",*chs,rdc&0xff)) ;
	}
      }
    }
  }
  if (status>=0) status=n;
  SERDEBUGPR(0x2,("***%d***\n",status));
  return status;
}

#define RCV_BUFF 256

#if !HAVE_SELECT
static void
alrm(junk)
int junk;
{
    /* do nothing */
    signal(SIGALRM, alrm);
}
#endif

/* arguments for call to readbyte() */
#if HAVE_SELECT
#  define RB_POLL   0
#  define RB_BLOCK -1
#else
#  define RB_POLL   1
#  define RB_BLOCK  0
#endif

static int test_carrierdetect(serial_port_info *port) {
  int modeminfo=0;
#if !defined(__CYGWIN__) && !defined(DBOX2)
  if (ioctl (port->fd,TIOCMGET,&modeminfo) < 0) {    errprint(errno,"ioctl(serial_port, TIOCMGET) failed");
    exit(1);
  }
  if ((modeminfo &= port->reset_detect_mask)) {
    if (!have_reset) SERDEBUGPR(0xfffe,(" **test_reset()** "));
    have_reset=1;
  }
#endif

#if defined(DBOX2)
  if(port->rtsctsfd > 0) {
	  unsigned int status;
	  ioctl(port->rtsctsfd, GET_PCDAT, &status);
	  if(((status >> 10) & 0x01)) {
		  modeminfo |= TIOCM_CAR;
		  have_reset=1;
	  }
  }
#endif
  return modeminfo ;
}

/*
  bytes are received standard from serial port, no inverting
  og shift of bits
*/
static int rcv_count=0;

/*      Throw away all input characters until no more are sent. */
void  purge(serial_port_info *port, short skiploop)
{
#if _POSIX_SOURCE || DBOX2
        if (tcflush(port->fd, TCIOFLUSH  ) < 0)
	  errprint(errno," tcflush on serial port failed\n");
#else
        if (!skiploop)
	  while (readbyte_norm(0,1000,0) != -4)
                ; /* empty loop */
#endif
        rcv_count = 0;
}

static int check_and_handle_reset(int flags) {
  if (port->flags&RESET_DETECT && flags&CHKRESET) {
    int cd=have_reset;
    if (!have_reset)  cd=test_carrierdetect(port);
    if (cd) {
      /* SERDEBUGPR(0xffff,(" **check&handel_reset()** ")); */
      /* while(!test_carrierdetect(port)) micro_sleep(10); */
      purge(port,1); /* flush buffer & input */
      return 1;
    }
  }
  return 0;
}


int readbyte_port(int deb, long useconds, int flags ) {
/*
  returns
    >=0   : bytevalue read
    -1    : reset detected (carrier_detect off)
    -2    : read port error
    -3    : error in select
 */
  int i=-4;
  if (port->fd<0) return -2;

  while (i== -4){
    i= readbyte_norm(port,useconds,flags);
  }

  flags &= SERDB_RD|SERDB_WR|SERDB_RDWR;
  if (i<0) {
	SERDEBUGPR(flags,("[r:%d]",i));
  } else {
    i=  swapbits(i); /* 0xff & (port->bytereverse ? ~swapbits(i) : i ); */
    SERDEBUGPR(flags,("[r:%02x]",i));
  }
  return i;
}

int readbyte_rx(int deb,  int flags ) {
  int r;
  r=readbyte_port(deb, serial_delay.rd, flags);
  if (deb) mlogprint(deb," %02x", (r&0xff));
  return r;
}


/*
  bytes are received standard from serial port, no inverting
  og shift of bits
*/
int readbyte_norm(serial_port_info  *port, long useconds, int flags) {
/*
  returns
    >=0   : bytevalue read
    -1    : reset detected
    -2    : read port error
    -3    : error in select
    -4    : no data
 */
#if HAVE_SELECT
  struct timeval tmval = { 0L, 0L }, *tmval_p ;
  fd_set refds ;
  int srval ;
#endif
  static char *p, rxbuf[RCV_BUFF];

  if (check_and_handle_reset(flags)) return -1;

  if (rcv_count > 0){
    rcv_count--;
    return(*p++ & 0xff);
  }


#if HAVE_SELECT
  FD_ZERO(&refds);
  FD_SET(port->fd,&refds);
  if(useconds == RB_BLOCK) { /* block */
    tmval_p = NULL ;
  } else {
    tmval_p = &tmval ;
    tmval.tv_sec = (unsigned long) useconds/1000000L ;
    tmval.tv_usec = (unsigned long) (useconds%1000000L)+1 ;
    SERDEBUGPR(0x8,("%lu:%lu ",tmval.tv_sec,tmval.tv_usec));
  }

  srval = select(port->fd+2, &refds, NULL, NULL, tmval_p) ;
  if ( srval < 0 ){
    errprint(errno," select() error in %s\r\n", __FILE__ );
    return -3 ;
  }
  if (check_and_handle_reset(flags)) return -1;
  if ( srval == 0 ) {
    SERDEBUGPR(0x1000,("."));
    return -4 ;
  }
#if 0
  /* equivalent to srval==0  ::*/
  if( ! FD_ISSET(port->fd, &refds) ) {/* nothing to do, ergo return */
    SERDEBUGPR(0x1000,("."));
    return -4 ;
  }
#endif

  rcv_count = read(port->fd, p = rxbuf, RCV_BUFF);

#else
  serfunc_is_not_tested_without_select();
  if (useconds){
    signal(SIGALRM, alrm);
    alarm((unsigned) useconds);
  }
  if ((rcv_count = read(port->fd, p = rxbuf, RCV_BUFF)) < 1)
    return(-2);
  if (mseconds)
    alarm(0);
#endif
  rcv_count--;
  return(*p++ & 0xff);
}

/* base+5: Line status register b7 b6....:
   b7= 0
   b6= TSR empty
   b5= THR empty
   b4=_Break interupt indicator
   b3= framing error
   b2= parity error
   b1= overrun error
   b0= receive data ready
*/
/* base+6: modem status register
   b7= -RLSD   TIOCM_CAR = TIOCM_CD
   b6= -RI     TIOCM_RI
   b5= -DSR    TIOCM_DSR
   b4= -CTS    TIOCM_CTS
   b3= Delta RLSD  ( 1= -RLSD has changed state)
   b2= trailing edge ring indicator (1= -RI has changed state)
   b1= Delta DSR indicator (1= -DSR has changed state )
   b0= Delta CTS indicator (1= -CTS has changed state )

Only connected in season interface:
   Rx-Data, Tx-Data, Gnd, and CD/Reset.

*/

int GetReset_x(int deb) {
  int cd;
  if (port->fd<0) return 0;
  have_reset=0;
  if (test_carrierdetect(port)) SERDEBUGPR(0xffe,(" **Get_Reset()**\n"));
  purge(port,0);
  have_reset=0;
  SERDEBUGPR(0x1,("\n"));
  /* wait for reset goes off ?? */
  if (port->flags&RESET_DETECT) {
    SERDEBUGPR(0xffe,(" **test_carrierdetect loop:** "));
    while(test_carrierdetect(port) ) {
      if (serial_delay.reset_loop) micro_sleep(serial_delay.reset_loop);
      have_reset=0;
      /* dummya = (dummya*17)%11; */
    }
  }
  if (serial_delay.reset_off) micro_sleep(serial_delay.reset_off);
  /* data must start in maximum/minimum? 10 mikroseconds after
    reset is off ! */
  /* SERDEBUGPR(0x1,("carrier on\n"));
     micro_sleep(2);
     SERDEBUGPR(0xf,("reset off\n")); */
  have_reset=0;
  cd=test_carrierdetect(port);
  return !cd;
}

void openPort_season(int comport, char *ldevname) {
  /* set device name, if *devname empty/null use comport[1..4] */
  if (ldevname ? strlen(ldevname)>0 : 0) {
    devname=ldevname;
  } else {
    devttySn[9] = comport-1+'0';
    devname= devttySn;
  }
  port->flags = RESET_DETECT;
}

void openPort_phoenix(int comport, char *ldevname) {
  /* set device name, if *devname empty/null use comport[1..4] */
  if (ldevname ? strlen(ldevname)>0 : 0) {
    devname=ldevname;
  } else {
    devttySn[9] = comport-1+'0';
    devname= devttySn;
  }

  port->flags = RESET_DETECT;

}

int initPort_season(unsigned char csize, unsigned char parity, unsigned char cstopb,
	      unsigned char byteconvention, int baud ) {
  /* csize: 5/6/7/8 ; parity: 0=none, 1=odd, 2=even ; cstopb=1/2 */

  int portno=0;
  int portres=0;

#ifdef USE_PRIORITY_SCHEDULING
  /*
   * enable real-time scheduling so we can use short delays with nanosleep
   */
#ifdef _POSIX_PRIORITY_SCHEDULING
  int ern;
  struct sched_param mysched;
  mysched.sched_priority = 90;
  ern=sched_setscheduler(0, SCHED_FIFO, &mysched);
  if (ern ) SERDEBUGPR(0xffff,(" sched_setschedular(0,SCHED_FIFO,..) returned: %d, so scheduler is not changed\n",ern ));
#endif
#endif

  init_bits4();
  my_port[portno].bytereverse = byteconvention;
  my_port[portno].serparity = parity? (parity&1 ? PARENB|PARODD : PARENB) : 0;
  my_port[portno].cstopb = cstopb==1 ? 0 : CSTOPB ;
  switch (csize) {
  case 5:  my_port[portno].cbitsize = CS5; break;
  case 6:  my_port[portno].cbitsize = CS6; break;
  case 7:  my_port[portno].cbitsize = CS7; break;
  default:  my_port[portno].cbitsize = CS8; break;
  }

  if (open_serial_port_season(portno,devname)<0) {
      do_port_error("open port error\n");
      portres=-4;
  } else {
      /* port_fprintf_info(stdout,portno); */
      set_bps(portno,baud);
      portres=port_fprintf_info(stdout,portno);
  }
  port->flags = RESET_DETECT;
  return portres;
}

int initPort_phoenix(unsigned char csize, unsigned char parity, unsigned char cstopb,
	      unsigned char byteconvention, int baud ) {
  /* csize: 5/6/7/8 ; parity: 0=none, 1=odd, 2=even ; cstopb=1/2 */

  int portno=0;
  int portres=0;

#ifdef USE_PRIORITY_SCHEDULING
  /*
   * enable real-time scheduling so we can use short delays with nanosleep
   */
#ifdef _POSIX_PRIORITY_SCHEDULING
  int ern;
  struct sched_param mysched;
  mysched.sched_priority = 90;
  ern=sched_setscheduler(0, SCHED_FIFO, &mysched);
  if (ern ) SERDEBUGPR(0xffff,(" sched_setschedular(0,SCHED_FIFO,..) returned: %d, so scheduler is not changed\n",ern ));
#endif
#endif

  init_bits4();
  my_port[portno].bytereverse = byteconvention;
  my_port[portno].serparity = parity? (parity&1 ? PARENB|PARODD : PARENB) : 0;
  my_port[portno].cstopb = cstopb==1 ? 0 : CSTOPB ;
  switch (csize) {
  case 5:  my_port[portno].cbitsize = CS5; break;
  case 6:  my_port[portno].cbitsize = CS6; break;
  case 7:  my_port[portno].cbitsize = CS7; break;
  default:  my_port[portno].cbitsize = CS8; break;
  }

  if (open_serial_port_phoenix(portno,devname)<0) {
      do_port_error("open port error\n");
      portres=-4;
  } else {
      /* port_fprintf_info(stdout,portno); */
      set_bps(portno,baud);
      portres=port_fprintf_info(stdout,portno);
  }

  port->flags = RESET_DETECT;

  return portres;
}

/* ToggleRTS
   Handles the RTS line toggle causing a smartcard to reset */
void ToggleRTS(int deb)
{
  int status;
  ioctl(deb, TIOCMGET, &status);

  SERDEBUGPR(0xfe,("Toggle RTS: "));
  if(status & TIOCM_RTS) {
    status &= ~TIOCM_RTS;
    SERDEBUGPR(0xfe,("RTS off\n"));
  } else {
    status |= TIOCM_RTS;
    SERDEBUGPR(0xfe,("RTS on\n"));
  }
  ioctl(deb,TIOCMSET,&status);
}

/* Reset
   Handles the resetting of ISO7816 compliant smartcards */
int phoenix_reset(int deb)
{
  int r;
  SERDEBUGPR(0xfe,("Reset: resetting card...\n"));
  tcflush(deb,TCIOFLUSH);
  ToggleRTS(deb);
  usleep(40*1000);
  ToggleRTS(deb);
  r=readbyte_norm(0, 2000, 0);
  if(r>0) {
    SERDEBUGPR(0xfe,("Reset: ATR len=%02d\n",r));
  }
  return r;
}

